from collections.abc import Callable
from typing import Literal

from torch import Tensor, nn
from torch.nn.functional import relu

__all__ = [
    "wideresnet28x10",
]


class WideBasicBlock(nn.Module):
    def __init__(
        self,
        in_planes: int,
        planes: int,
        dropout_rate: float,
        stride: int,
        groups: int,
        conv_bias: bool,
        activation_fn: Callable,
        normalization_layer: type[nn.Module],
    ) -> None:
        super().__init__()
        self.activation_fn = activation_fn
        self.conv1 = nn.Conv2d(
            in_planes,
            planes,
            kernel_size=3,
            padding=1,
            groups=groups,
            bias=conv_bias,
        )
        self.dropout = nn.Dropout2d(p=dropout_rate)
        self.bn1 = normalization_layer(planes)
        self.conv2 = nn.Conv2d(
            planes,
            planes,
            kernel_size=3,
            stride=stride,
            padding=1,
            groups=groups,
            bias=conv_bias,
        )
        self.bn2 = normalization_layer(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(
                    in_planes,
                    planes,
                    kernel_size=1,
                    stride=stride,
                    groups=groups,
                    bias=conv_bias,
                ),
            )

    def forward(self, x: Tensor) -> Tensor:
        out = self.activation_fn(self.bn1(self.dropout(self.conv1(x))))
        out = self.conv2(out)
        out += self.shortcut(x)
        return self.activation_fn(self.bn2(out))


class _WideResNet(nn.Module):
    def __init__(
        self,
        depth: int,
        widen_factor: int,
        in_channels: int,
        num_classes: int,
        conv_bias: bool,
        dropout_rate: float,
        groups: int = 1,
        style: Literal["imagenet", "cifar"] = "imagenet",
        activation_fn: Callable = relu,
        normalization_layer: type[nn.Module] = nn.BatchNorm2d,
    ) -> None:
        super().__init__()
        self.dropout_rate = dropout_rate
        self.activation_fn = activation_fn
        self.in_planes = 16

        if (depth - 4) % 6 != 0:
            raise ValueError(f"Wide-resnet depth should be 6n+4. Got {depth}.")
        num_blocks = int((depth - 4) / 6)
        num_stages = [
            16,
            16 * widen_factor,
            32 * widen_factor,
            64 * widen_factor,
        ]

        if style == "imagenet":
            self.conv1 = nn.Conv2d(
                in_channels,
                num_stages[0],
                kernel_size=7,
                stride=2,
                padding=3,
                groups=groups,
                bias=conv_bias,
            )
        elif style == "cifar":
            self.conv1 = nn.Conv2d(
                in_channels,
                num_stages[0],
                kernel_size=3,
                stride=1,
                padding=1,
                groups=groups,
                bias=conv_bias,
            )
        else:
            raise ValueError(f"Unknown WideResNet style: {style}. ")

        self.bn1 = normalization_layer(num_stages[0])

        if style == "imagenet":
            self.optional_pool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        else:
            self.optional_pool = nn.Identity()

        self.layer1 = self._wide_layer(
            WideBasicBlock,
            num_stages[1],
            num_blocks=num_blocks,
            dropout_rate=dropout_rate,
            stride=1,
            groups=groups,
            activation_fn=activation_fn,
            conv_bias=conv_bias,
            normalization_layer=normalization_layer,
        )
        self.layer2 = self._wide_layer(
            WideBasicBlock,
            num_stages[2],
            num_blocks=num_blocks,
            dropout_rate=dropout_rate,
            stride=2,
            groups=groups,
            activation_fn=activation_fn,
            conv_bias=conv_bias,
            normalization_layer=normalization_layer,
        )
        self.layer3 = self._wide_layer(
            WideBasicBlock,
            num_stages[3],
            num_blocks=num_blocks,
            dropout_rate=dropout_rate,
            stride=2,
            groups=groups,
            activation_fn=activation_fn,
            conv_bias=conv_bias,
            normalization_layer=normalization_layer,
        )
        self.final_dropout = nn.Dropout(p=dropout_rate)
        self.pool = nn.AdaptiveAvgPool2d(output_size=1)
        self.flatten = nn.Flatten(1)
        self.linear = nn.Linear(
            num_stages[3],
            num_classes,
        )

    def _wide_layer(
        self,
        block: type[WideBasicBlock],
        planes: int,
        num_blocks: int,
        dropout_rate: float,
        stride: int,
        groups: int,
        conv_bias: bool,
        activation_fn: Callable,
        normalization_layer: type[nn.Module],
    ) -> nn.Module:
        strides = [stride] + [1] * (int(num_blocks) - 1)
        layers = []

        for stride in strides:
            layers.append(
                block(
                    in_planes=self.in_planes,
                    planes=planes,
                    stride=stride,
                    dropout_rate=dropout_rate,
                    groups=groups,
                    conv_bias=conv_bias,
                    activation_fn=activation_fn,
                    normalization_layer=normalization_layer,
                )
            )
            self.in_planes = planes
        return nn.Sequential(*layers)

    def feats_forward(self, x: Tensor) -> Tensor:
        out = self.activation_fn(self.bn1(self.conv1(x)))
        out = self.optional_pool(out)
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.pool(out)
        return self.final_dropout(self.flatten(out))

    def forward(self, x: Tensor) -> Tensor:
        return self.linear(self.feats_forward(x))


def wideresnet28x10(
    in_channels: int,
    num_classes: int,
    conv_bias: bool = True,
    dropout_rate: float = 0.3,
    groups: int = 1,
    style: Literal["imagenet", "cifar"] = "imagenet",
    activation_fn: Callable = relu,
    normalization_layer: type[nn.Module] = nn.BatchNorm2d,
) -> _WideResNet:
    """Wide-ResNet-28x10 from `Wide Residual Networks
    <https://arxiv.org/pdf/1605.07146.pdf>`_.

    Args:
        in_channels (int): Number of input channels
        num_classes (int): Number of classes to predict.
        groups (int, optional): Number of groups in convolutions. Defaults to
            ``1``.
        conv_bias (bool): Whether to use bias in convolutions. Defaults to
            ``True``.
        dropout_rate (float, optional): Dropout rate. Defaults to ``0.3``.
        style (bool, optional): Whether to use the ImageNet
            structure. Defaults to ``True``.
        activation_fn (Callable, optional): Activation function. Defaults to
            ``torch.nn.functional.relu``.
        normalization_layer (nn.Module, optional): Normalization layer.
            Defaults to ``torch.nn.BatchNorm2d``.

    Returns:
        _Wide: A Wide-ResNet-28x10.
    """
    return _WideResNet(
        depth=28,
        widen_factor=10,
        in_channels=in_channels,
        conv_bias=conv_bias,
        dropout_rate=dropout_rate,
        num_classes=num_classes,
        groups=groups,
        style=style,
        activation_fn=activation_fn,
        normalization_layer=normalization_layer,
    )



"""
Improved Ensemble parameter-efficiency with Packed-Ensembles
============================================================

*This tutorial is adapted from a notebook part of a lecture given at the `Helmholtz AI Conference <https://haicon24.de/>`_ by Sebastian Starke, Peter Steinbach, Gianni Franchi, and Olivier Laurent.*

In this notebook will work on the MNIST dataset that was introduced by Corinna Cortes, Christopher J.C. Burges, and later modified by Yann LeCun in the foundational paper:

- `Y. LeCun, L. Bottou, Y. Bengio, and P. Haffner. "Gradient-based learning applied to document recognition." Proceedings of the IEEE. <http://yann.lecun.com/exdb/publis/pdf/lecun-98.pdf>`_.

The MNIST dataset consists of 70 000 images of handwritten digits from 0 to 9. The images are grayscale and 28x28-pixel sized. The task is to classify the images into their respective digits. The dataset can be automatically downloaded using the `torchvision` library.

In this notebook, we will train a model and an ensemble on this task and evaluate their performance. The performance will consist in the following metrics:
- Accuracy: the proportion of correctly classified images,
- Brier score: a measure of the quality of the predicted probabilities,
- Calibration error: a measure of the calibration of the predicted probabilities,
- Negative Log-Likelihood: the value of the loss on the test set.

Throughout this notebook, we abstract the training and evaluation process using `PyTorch Lightning <https://lightning.ai/docs/pytorch/stable/>`_
and `TorchUncertainty <https://torch-uncertainty.github.io/>`_.

Similarly to keras for tensorflow, PyTorch Lightning is a high-level interface for PyTorch that simplifies the training and evaluation process using a Trainer.
TorchUncertainty is partly built on top of PyTorch Lightning and provides tools to train and evaluate models with uncertainty quantification.

TorchUncertainty includes datamodules that handle the data loading and preprocessing. We don't use them here for tutorial purposes.

1. Download, instantiate and visualize the datasets
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The dataset is automatically downloaded using torchvision. We then visualize a few images to see a bit what we are working with.
"""
# %%
import torch
import torchvision.transforms as T

# We set the number of epochs to some low value for the sake of time
max_epochs = 2

# Create the transforms for the images
train_transform = T.Compose(
    [
        T.ToTensor(),
        # We perform random cropping as data augmentation
        T.RandomCrop(28, padding=4),
        # As for the MNIST1d dataset, we normalize the data
        T.Normalize((0.1307,), (0.3081,)),
    ]
)
test_transform = T.Compose(
    [
        T.Grayscale(num_output_channels=1),
        T.ToTensor(),
        T.CenterCrop(28),
        T.Normalize((0.1307,), (0.3081,)),
    ]
)

# Download and instantiate the dataset
from torch.utils.data import Subset
from torchvision.datasets import MNIST, FashionMNIST

train_data = MNIST(
    root="./data/", download=True, train=True, transform=train_transform
)
test_data = MNIST(root="./data/", train=False, transform=test_transform)
# We only take the first 10k images to have the same number of samples as the test set using torch Subsets
ood_data = Subset(
    FashionMNIST(root="./data/", download=True, transform=test_transform),
    indices=range(10000),
)

# Create the corresponding dataloaders
from torch.utils.data import DataLoader

train_dl = DataLoader(train_data, batch_size=32, shuffle=True)
test_dl = DataLoader(test_data, batch_size=32, shuffle=False)
ood_dl = DataLoader(ood_data, batch_size=32, shuffle=False)

# %%
# You could replace all this cell by simply loading the MNIST datamodule from TorchUncertainty.
# Now, let's visualize a few images from the dataset. For this task, we use the viz_data dataset that applies no transformation to the images.

# Datasets without transformation to visualize the unchanged data
viz_data = MNIST(root="./data/", train=False)
ood_viz_data = FashionMNIST(root="./data/", download=True)

print("In distribution data:")
viz_data[0][0]
# %%
print("Out of distribution data:")
ood_viz_data[0][0]

# %%
# 2. Create & train the model
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# We will create a simple convolutional neural network (CNN): the LeNet model (also introduced by LeCun).
import torch.nn as nn
import torch.nn.functional as F


class LeNet(nn.Module):
    def __init__(
        self,
        in_channels: int,
        num_classes: int,
    ) -> None:
        super().__init__()
        self.conv1 = nn.Conv2d(in_channels, 6, (5, 5))
        self.conv2 = nn.Conv2d(6, 16, (5, 5))
        self.pooling = nn.AdaptiveAvgPool2d((4, 4))
        self.fc1 = nn.Linear(256, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, num_classes)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = F.relu(self.conv1(x))
        out = F.max_pool2d(out, 2)
        out = F.relu(self.conv2(out))
        out = F.max_pool2d(out, 2)
        out = torch.flatten(out, 1)
        out = F.relu(self.fc1(out))
        out = F.relu(self.fc2(out))
        return self.fc3(out)  # No softmax in the model!


# Instantiate the model, the images are in grayscale so the number of channels is 1
model = LeNet(in_channels=1, num_classes=10)

# %%
# We now need to define the optimization recipe:
# - the optimizer, here the standard stochastic gradient descent (SGD) with a learning rate of 0.05
# - the scheduler, here cosine annealing.


def optim_recipe(model, lr_mult: float = 1.0):
    optimizer = torch.optim.SGD(model.parameters(), lr=0.05 * lr_mult)
    scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=10)
    return {"optimizer": optimizer, "scheduler": scheduler}


# %%
# To train the model, we use `TorchUncertainty <https://torch-uncertainty.github.io/>`_, a library that we have developed to ease
# the training and evaluation of models with uncertainty.
#
# **Note:** To train supervised classification models we most often use the cross-entropy loss.
# With weight-decay, minimizing this loss amounts to finding a Maximum a posteriori (MAP) estimate of the model parameters.
# This means that the model is trained to predict the most likely class for each input.


from torch_uncertainty.routines import ClassificationRoutine
from torch_uncertainty import TUTrainer

# Create the trainer that will handle the training
trainer = TUTrainer(accelerator="cpu", max_epochs=max_epochs)

# The routine is a wrapper of the model that contains the training logic with the metrics, etc
routine = ClassificationRoutine(
    num_classes=10,
    model=model,
    loss=nn.CrossEntropyLoss(),
    optim_recipe=optim_recipe(model),
    eval_ood=True,
)

# In practice, avoid performing the validation on the test set (if you do model selection)
trainer.fit(routine, train_dataloaders=train_dl, val_dataloaders=test_dl)

# %%
# Evaluate the trained model on the test set - pay attention to the cls/Acc metric
perf = trainer.test(routine, dataloaders=[test_dl, ood_dl])

# %%
# This table provides a lot of information:
#
# **OOD Detection: Binary Classification MNIST vs. FashionMNIST**
# - AUPR/AUROC/FPR95: Measures the quality of the OOD detection. The higher the better for AUPR and AUROC, the lower the better for FPR95.
#
# **Calibration: Reliability of the Predictions**
# - ECE: Expected Calibration Error. The lower the better.
# - aECE: Adaptive Expected Calibration Error. The lower the better. (~More precise version of the ECE)
#
# **Classification Performance**
# - Accuracy: The ratio of correctly classified images. The higher the better.
# - Brier: The quality of the predicted probabilities (Mean Squared Error of the predictions vs. ground-truth). The lower the better.
# - Negative Log-Likelihood: The value of the loss on the test set. The lower the better.
#
# **Selective Classification & Grouping Loss**
# - We talk about these points later in the "To go further" section.
#
# 3. Training an ensemble of models with TorchUncertainty
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# You have two options here, you can either train the ensemble directly if you have enough memory,
# otherwise, you can train independent models and do the ensembling during the evaluation (sometimes called inference).
#
# In this case, we will do it sequentially. In this tutorial, you have the choice between training multiple models,
# which will take time if you have no GPU, or downloading the pre-trained models that we have prepared for you.
#
# Training the ensemble
#
# To train the ensemble, you will have to use the "deep_ensembles" function from TorchUncertainty, which will
# replicate and change the initialization of your networks to ensure diversity.

from torch_uncertainty.models import deep_ensembles
from torch_uncertainty.transforms import RepeatTarget

# Create the ensemble model
ensemble = deep_ensembles(
    LeNet(in_channels=1, num_classes=10),
    num_estimators=2,
    task="classification",
    reset_model_parameters=True,
)

trainer = TUTrainer(accelerator="cpu", max_epochs=1)
ens_routine = ClassificationRoutine(
    is_ensemble=True,
    num_classes=10,
    model=ensemble,
    loss=nn.CrossEntropyLoss(),  # The loss for the training
    format_batch_fn=RepeatTarget(
        2
    ),  # How to handle the targets when comparing the predictions
    optim_recipe=optim_recipe(
        ensemble, 2.0
    ),  # The optimization scheme with the optimizer and the scheduler as a dictionnary
    eval_ood=True,  # We want to evaluate the OOD-related metrics
)
trainer.fit(ens_routine, train_dataloaders=train_dl, val_dataloaders=test_dl)
ens_perf = trainer.test(ens_routine, dataloaders=[test_dl, ood_dl])

# %%
# The results are not comparable since we only trained the ensemble for one epoch to reduce GitHub's cpu usage.
# Feel free to run the notebook on your machine for a longer duration.
#
# We need to multiply the learning rate by 2 to account for the fact that we have 4 models
# in the ensemble and that we average the loss over all the predictions.
#
# #### Downloading the pre-trained models
#
# We have put the pre-trained models on Hugging Face that you can download with the utility function
# "hf_hub_download" imported just below. These models are trained for 75 epochs and are therefore not
# comparable to the all the other models trained in this notebook. The pretrained models can be seen
# on `HuggingFace <https://huggingface.co/ENSTA-U2IS/tutorial-models>`_ and TorchUncertainty's are `there <https://huggingface.co/torch-uncertainty>`_.

from torch_uncertainty.utils.hub import hf_hub_download

all_models = []
for i in range(8):
    hf_hub_download(
        repo_id="ENSTA-U2IS/tutorial-models",
        filename=f"version_{i}.ckpt",
        local_dir="./models/",
    )
    model = LeNet(in_channels=1, num_classes=10)
    state_dict = torch.load(f"./models/version_{i}.ckpt", map_location="cpu")[
        "state_dict"
    ]
    state_dict = {k.replace("model.", ""): v for k, v in state_dict.items()}
    model.load_state_dict(state_dict)
    all_models.append(model)

from torch_uncertainty.models import deep_ensembles
from torch_uncertainty.transforms import RepeatTarget

ensemble = deep_ensembles(
    all_models,
    num_estimators=None,
    task="classification",
    reset_model_parameters=True,
)

ens_routine = ClassificationRoutine(
    is_ensemble=True,
    num_classes=10,
    model=ensemble,
    loss=nn.CrossEntropyLoss(),  # The loss for the training
    format_batch_fn=RepeatTarget(
        8
    ),  # How to handle the targets when comparing the predictions
    optim_recipe=None,  # No optim recipe as the model is already trained
    eval_ood=True,  # We want to evaluate the OOD-related metrics
)

trainer = TUTrainer(accelerator="cpu", max_epochs=max_epochs)

ens_perf = trainer.test(ens_routine, dataloaders=[test_dl, ood_dl])

# %%
# 4. From Deep Ensembles to Packed-Ensembles
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# In the paper `Packed-Ensembles for Efficient Uncertainty Quantification <https://arxiv.org/abs/2210.09184>`_
# published at the International Conference on Learning Representations (ICLR) in 2023, we introduced a
# modification of Deep Ensembles to make it more computationally-efficient. The idea is to pack the ensemble
# members into a single model, which allows us to train the ensemble in a single forward pass.
# This modification is particularly useful when the ensemble size is large, as it is often the case in practice.
#
# We will need to update the model and replace the layers with their Packed equivalents. You can find the
# documentation of the Packed-Linear layer using this `link <https://torch-uncertainty.github.io/generated/torch_uncertainty.layers.PackedLinear.html>`_,
# and the Packed-Conv2D, `here <https://torch-uncertainty.github.io/generated/torch_uncertainty.layers.PackedLinear.html>`_.

import torch
import torch.nn as nn
from einops import rearrange

from torch_uncertainty.layers import PackedConv2d, PackedLinear


class PackedLeNet(nn.Module):
    def __init__(
        self,
        in_channels: int,
        num_classes: int,
        alpha: int,
        num_estimators: int,
    ) -> None:
        super().__init__()
        self.num_estimators = num_estimators
        self.conv1 = PackedConv2d(
            in_channels,
            6,
            (5, 5),
            alpha=alpha,
            num_estimators=num_estimators,
            first=True,
        )
        self.conv2 = PackedConv2d(
            6,
            16,
            (5, 5),
            alpha=alpha,
            num_estimators=num_estimators,
        )
        self.pooling = nn.AdaptiveAvgPool2d((4, 4))
        self.fc1 = PackedLinear(
            256, 120, alpha=alpha, num_estimators=num_estimators
        )
        self.fc2 = PackedLinear(
            120, 84, alpha=alpha, num_estimators=num_estimators
        )
        self.fc3 = PackedLinear(
            84,
            num_classes,
            alpha=alpha,
            num_estimators=num_estimators,
            last=True,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = F.relu(self.conv1(x))
        out = F.max_pool2d(out, 2)
        out = F.relu(self.conv2(out))
        out = F.max_pool2d(out, 2)
        out = rearrange(
            out, "e (m c) h w -> (m e) c h w", m=self.num_estimators
        )
        out = torch.flatten(out, 1)
        out = F.relu(self.fc1(out))
        out = F.relu(self.fc2(out))
        return self.fc3(out)  # Again, no softmax in the model


# Instantiate the model, the images are in grayscale so the number of channels is 1
packed_model = PackedLeNet(
    in_channels=1, num_classes=10, alpha=2, num_estimators=4
)

# Create the trainer that will handle the training
trainer = TUTrainer(accelerator="cpu", max_epochs=max_epochs)

# The routine is a wrapper of the model that contains the training logic with the metrics, etc
packed_routine = ClassificationRoutine(
    is_ensemble=True,
    num_classes=10,
    model=packed_model,
    loss=nn.CrossEntropyLoss(),
    format_batch_fn=RepeatTarget(4),
    optim_recipe=optim_recipe(packed_model, 4.0),
    eval_ood=True,
)

# In practice, avoid performing the validation on the test set
trainer.fit(packed_routine, train_dataloaders=train_dl, val_dataloaders=test_dl)

packed_perf = trainer.test(packed_routine, dataloaders=[test_dl, ood_dl])

# %%
# The training time should be approximately similar to the one of the single model that you trained before. However, please note that we are working with very small models, hence completely underusing your GPU. As such, the training time is not representative of what you would observe with larger models.
#
# You can read more on Packed-Ensembles in the `paper <https://arxiv.org/abs/2210.09184>`_ or the `Medium <https://medium.com/@adrien.lafage/make-your-neural-networks-more-reliable-with-packed-ensembles-7ad0b737a873>`_ post.
#
# To Go Further & More Concepts of Uncertainty in ML
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# **Question 1:** Have a look at the models in the "lightning_logs". If you are on your own machine, try to visualize the learning curves with `tensorboard --logdir lightning_logs`.
#
# **Question 2:** Add a cell below and try to find the errors made by packed-ensembles on the test set. Visualize the errors and their labels and look at the predictions of the different sub-models. Are they similar? Can you think of uncertainty scores that could help you identify these errors?
#
# Selective Classification
# ^^^^^^^^^^^^^^^^^^^^^^^^
#
# Selective classification or "prediction with rejection" is a paradigm in uncertainty-aware machine learning where the model can decide not to make a prediction if the confidence score given by the model is below some pre-computed threshold. This can be useful in real-world applications where the cost of making a wrong prediction is high.
#
# In constrast to calibration, the values of the confidence scores are not important, only the order of the scores. *Ideally, the best model will order all the correct predictions first, and all the incorrect predictions last.* In this case, there will be a threshold so that all the predictions above the threshold are correct, and all the predictions below the threshold are incorrect.
#
# In TorchUncertainty, we look at 3 different metrics for selective classification:
# - **AURC**: The area under the Risk (% of errors) vs. Coverage (% of classified samples) curve. This curve expresses how the risk of the model evolves as we increase the coverage (the proportion of predictions that are above the selection threshold). This metric will be minimized by a model able to perfectly separate the correct and incorrect predictions.
#
# The following metrics are computed at a fixed risk and coverage level and that have practical interests. The idea of these metrics is that you can set the selection threshold to achieve a certain level of risk and coverage, as required by the technical constraints of your application:
# - **Coverage at 5% Risk**: The proportion of predictions that are above the selection threshold when it is set for the risk to egal 5%. Set the risk threshold to your application constraints. The higher the better.
# - **Risk at 80% Coverage**: The proportion of errors when the coverage is set to 80%. Set the coverage threshold to your application constraints. The lower the better.
#
# Grouping Loss
# ^^^^^^^^^^^^^
#
# The grouping loss is a measure of uncertainty orthogonal to calibration. Have a look at `this paper <https://arxiv.org/abs/2210.16315>`_ to learn about it. Check out their small library `GLest <https://github.com/aperezlebel/glest>`_. TorchUncertainty includes a wrapper of the library to compute the grouping loss with eval_grouping_loss parameter.



import pytest
import torch

from torch_uncertainty.transforms import Cutout


class TestCutout:
    """Testing the Cutout transform."""

    def test_cutout_allchannels(self):
        inputs = torch.rand(32, 32) + 10  # no zeros
        transform = Cutout(16)
        output = transform(inputs)

        assert (output == 0).sum() > 0

        inputs = torch.rand(1, 32, 32) + 10  # no zeros
        output = transform(inputs)

        assert (output == 0).sum() > 0

        inputs = torch.rand(3, 32, 32) + 10  # no zeros
        output = transform(inputs)

        assert (output == 0).sum() > 0

    def test_cutout_negative_length(self):
        with pytest.raises(ValueError):
            _ = Cutout(-1)

    def test_cutout_negative_value(self):
        with pytest.raises(ValueError):
            _ = Cutout(42, -16)



# ruff: noqa: F401
from .classification import ClassificationRoutine
from .pixel_regression import PixelRegressionRoutine
from .regression import RegressionRoutine
from .segmentation import SegmentationRoutine



from pathlib import Path
from typing import Literal

from torch_uncertainty.models import deep_ensembles
from torch_uncertainty.routines.classification import ClassificationRoutine
from torch_uncertainty.utils import get_version

from . import ResNetBaseline, VGGBaseline, WideResNetBaseline


class DeepEnsemblesBaseline(ClassificationRoutine):
    backbones = {
        "resnet": ResNetBaseline,
        "vgg": VGGBaseline,
        "wideresnet": WideResNetBaseline,
    }

    def __init__(
        self,
        num_classes: int,
        log_path: str | Path,
        checkpoint_ids: list[int],
        backbone: Literal["resnet", "vgg", "wideresnet"],
        eval_ood: bool = False,
        eval_shift: bool = False,
        eval_grouping_loss: bool = False,
        ood_criterion: Literal["msp", "logit", "energy", "entropy", "mi", "vr"] = "msp",
        log_plots: bool = False,
        calibration_set: Literal["val", "test"] = "val",
    ) -> None:
        log_path = Path(log_path)

        backbone_cls = self.backbones[backbone]

        models = []
        for version in checkpoint_ids:  # coverage: ignore
            ckpt_file, hparams_file = get_version(root=log_path, version=version)
            trained_model = backbone_cls.load_from_checkpoint(
                checkpoint_path=ckpt_file,
                hparams_file=hparams_file,
                loss=None,
                optim_recipe=None,
            ).eval()
            models.append(trained_model.model)
        de = deep_ensembles(models=models)

        super().__init__(  # coverage: ignore
            num_classes=num_classes,
            model=de,
            loss=None,
            is_ensemble=de.num_estimators > 1,
            eval_ood=eval_ood,
            eval_shift=eval_shift,
            eval_grouping_loss=eval_grouping_loss,
            ood_criterion=ood_criterion,
            log_plots=log_plots,
            calibration_set=calibration_set,
        )
        self.save_hyperparameters()  # coverage: ignore



from typing import Any, Literal

import torch.nn.functional as F
from einops import rearrange
from torch import Tensor, nn

from torch_uncertainty.layers import PackedConv2d, PackedLinear
from torch_uncertainty.utils import load_hf

from .utils import get_resnet_num_blocks

__all__ = [
    "packed_resnet",
]

weight_ids = {
    "10": {
        "18": "pe_resnet18_c10",
        "32": None,
        "50": "pe_resnet50_c10",
        "101": None,
        "152": None,
    },
    "100": {
        "18": "pe_resnet18_c100",
        "32": None,
        "50": "pe_resnet50_c100",
        "101": None,
        "152": None,
    },
    "1000": {
        "18": None,
        "32": None,
        "50": "pe_resnet50_in1k",
        "101": None,
        "152": None,
    },
}


class _BasicBlock(nn.Module):
    expansion = 1

    def __init__(
        self,
        in_planes: int,
        planes: int,
        stride: int,
        alpha: int,
        num_estimators: int,
        gamma: int,
        conv_bias: bool,
        dropout_rate: float,
        groups: int,
        normalization_layer: type[nn.Module],
    ) -> None:
        super().__init__()

        # No subgroups for the first layer
        self.conv1 = PackedConv2d(
            in_planes,
            planes,
            kernel_size=3,
            alpha=alpha,
            num_estimators=num_estimators,
            groups=groups,
            stride=stride,
            padding=1,
            bias=conv_bias,
        )
        self.bn1 = normalization_layer(planes * alpha)
        self.dropout = nn.Dropout2d(p=dropout_rate)
        self.conv2 = PackedConv2d(
            planes,
            planes,
            kernel_size=3,
            alpha=alpha,
            num_estimators=num_estimators,
            gamma=gamma,
            groups=groups,
            stride=1,
            padding=1,
            bias=conv_bias,
        )
        self.bn2 = normalization_layer(planes * alpha)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                PackedConv2d(
                    in_planes,
                    self.expansion * planes,
                    kernel_size=1,
                    alpha=alpha,
                    num_estimators=num_estimators,
                    gamma=gamma,
                    groups=groups,
                    stride=stride,
                    bias=conv_bias,
                ),
                normalization_layer(self.expansion * planes * alpha),
            )

    def forward(self, x: Tensor) -> Tensor:
        out = F.relu(self.dropout(self.bn1(self.conv1(x))))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        return F.relu(out)


class _Bottleneck(nn.Module):
    expansion = 4

    def __init__(
        self,
        in_planes: int,
        planes: int,
        stride: int,
        alpha: int,
        num_estimators: int,
        gamma: int,
        conv_bias: bool,
        dropout_rate: float,
        groups: int,
        normalization_layer: type[nn.Module],
    ) -> None:
        super().__init__()

        # No subgroups for the first layer
        self.conv1 = PackedConv2d(
            in_planes,
            planes,
            kernel_size=1,
            alpha=alpha,
            num_estimators=num_estimators,
            gamma=1,  # No groups from gamma in the first layer
            groups=groups,
            bias=conv_bias,
        )
        self.bn1 = normalization_layer(planes * alpha)
        self.conv2 = PackedConv2d(
            planes,
            planes,
            kernel_size=3,
            alpha=alpha,
            num_estimators=num_estimators,
            gamma=gamma,
            stride=stride,
            padding=1,
            groups=groups,
            bias=conv_bias,
        )
        self.bn2 = normalization_layer(planes * alpha)
        self.dropout = nn.Dropout2d(p=dropout_rate)
        self.conv3 = PackedConv2d(
            planes,
            self.expansion * planes,
            kernel_size=1,
            alpha=alpha,
            num_estimators=num_estimators,
            gamma=gamma,
            groups=groups,
            bias=conv_bias,
        )
        self.bn3 = normalization_layer(self.expansion * planes * alpha)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                PackedConv2d(
                    in_planes,
                    self.expansion * planes,
                    kernel_size=1,
                    alpha=alpha,
                    num_estimators=num_estimators,
                    gamma=gamma,
                    groups=groups,
                    stride=stride,
                    bias=conv_bias,
                ),
                normalization_layer(self.expansion * planes * alpha),
            )

    def forward(self, x: Tensor) -> Tensor:
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.dropout(self.bn2(self.conv2(out))))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        return F.relu(out)


class _PackedResNet(nn.Module):
    def __init__(
        self,
        block: type[_BasicBlock | _Bottleneck],
        num_blocks: list[int],
        in_channels: int,
        num_classes: int,
        conv_bias: bool,
        num_estimators: int,
        dropout_rate: float,
        alpha: int = 2,
        gamma: int = 1,
        groups: int = 1,
        style: Literal["imagenet", "cifar"] = "imagenet",
        in_planes: int = 64,
        normalization_layer: type[nn.Module] = nn.BatchNorm2d,
    ) -> None:
        super().__init__()

        self.in_channels = in_channels
        self.alpha = alpha
        self.gamma = gamma
        self.groups = groups
        self.num_estimators = num_estimators

        self.in_planes = in_planes
        block_planes = in_planes

        if style == "imagenet":
            self.conv1 = PackedConv2d(
                self.in_channels,
                block_planes,
                kernel_size=7,
                stride=2,
                padding=3,
                alpha=alpha,
                num_estimators=num_estimators,
                gamma=1,  # No groups for the first layer
                groups=groups,
                bias=conv_bias,
                first=True,
            )
        elif style == "cifar":
            self.conv1 = PackedConv2d(
                self.in_channels,
                block_planes,
                kernel_size=3,
                stride=1,
                padding=1,
                alpha=alpha,
                num_estimators=num_estimators,
                gamma=1,  # No groups for the first layer
                groups=groups,
                bias=conv_bias,
                first=True,
            )
        else:
            raise ValueError(f"Unknown style. Got {style}.")

        self.bn1 = normalization_layer(block_planes * alpha)

        if style == "imagenet":
            self.optional_pool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        else:
            self.optional_pool = nn.Identity()

        self.layer1 = self._make_layer(
            block,
            block_planes,
            num_blocks[0],
            stride=1,
            alpha=alpha,
            num_estimators=num_estimators,
            conv_bias=conv_bias,
            dropout_rate=dropout_rate,
            gamma=gamma,
            groups=groups,
            normalization_layer=normalization_layer,
        )
        self.layer2 = self._make_layer(
            block,
            block_planes * 2,
            num_blocks[1],
            stride=2,
            alpha=alpha,
            num_estimators=num_estimators,
            conv_bias=conv_bias,
            dropout_rate=dropout_rate,
            gamma=gamma,
            groups=groups,
            normalization_layer=normalization_layer,
        )
        self.layer3 = self._make_layer(
            block,
            block_planes * 4,
            num_blocks[2],
            stride=2,
            alpha=alpha,
            num_estimators=num_estimators,
            conv_bias=conv_bias,
            dropout_rate=dropout_rate,
            gamma=gamma,
            groups=groups,
            normalization_layer=normalization_layer,
        )
        if len(num_blocks) == 4:
            self.layer4 = self._make_layer(
                block,
                block_planes * 8,
                num_blocks[3],
                stride=2,
                alpha=alpha,
                num_estimators=num_estimators,
                conv_bias=conv_bias,
                dropout_rate=dropout_rate,
                gamma=gamma,
                groups=groups,
                normalization_layer=normalization_layer,
            )
            linear_multiplier = 8
        else:
            self.layer4 = nn.Identity()
            linear_multiplier = 4

        self.final_dropout = nn.Dropout(p=dropout_rate)
        self.pool = nn.AdaptiveAvgPool2d(output_size=1)
        self.flatten = nn.Flatten(1)

        self.linear = PackedLinear(
            block_planes * linear_multiplier * block.expansion,
            num_classes,
            alpha=alpha,
            num_estimators=num_estimators,
            last=True,
        )

    def _make_layer(
        self,
        block: type[_BasicBlock | _Bottleneck],
        planes: int,
        num_blocks: int,
        stride: int,
        alpha: int,
        num_estimators: int,
        conv_bias: bool,
        dropout_rate: float,
        gamma: int,
        groups: int,
        normalization_layer: type[nn.Module],
    ) -> nn.Module:
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(
                block(
                    in_planes=self.in_planes,
                    planes=planes,
                    stride=stride,
                    alpha=alpha,
                    num_estimators=num_estimators,
                    conv_bias=conv_bias,
                    dropout_rate=dropout_rate,
                    gamma=gamma,
                    groups=groups,
                    normalization_layer=normalization_layer,
                )
            )
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.optional_pool(out)
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)

        out = rearrange(out, "e (m c) h w -> (m e) c h w", m=self.num_estimators)

        out = self.pool(out)
        out = self.final_dropout(self.flatten(out))
        return self.linear(out)

    def check_config(self, config: dict[str, Any]) -> bool:
        """Check if the pretrained configuration matches the current model."""
        return (
            (config["alpha"] == self.alpha)
            * (config["gamma"] == self.gamma)
            * (config["groups"] == self.groups)
            * (config["num_estimators"] == self.num_estimators)
        )


def packed_resnet(
    in_channels: int,
    num_classes: int,
    arch: int,
    num_estimators: int,
    alpha: int,
    gamma: int,
    conv_bias: bool = True,
    width_multiplier: float = 1.0,
    groups: int = 1,
    dropout_rate: float = 0,
    style: Literal["imagenet", "cifar"] = "imagenet",
    normalization_layer: type[nn.Module] = nn.BatchNorm2d,
    pretrained: bool = False,
) -> _PackedResNet:
    """Packed-Ensembles of ResNet.

    Args:
        in_channels (int): Number of input channels.
        num_classes (int): Number of classes to predict.
        arch (int): The architecture of the ResNet.
        conv_bias (bool): Whether to use bias in convolutions. Defaults to
            ``True``.
        dropout_rate (float): Dropout rate. Defaults to 0.
        num_estimators (int): Number of estimators in the ensemble.
        alpha (int): Expansion factor affecting the width of the estimators.
        gamma (int): Number of groups within each estimator.
        width_multiplier (float): Width multiplier. Defaults to 1.
        groups (int): Number of groups within each estimator.
        style (bool, optional): Whether to use the ImageNet
            structure. Defaults to ``True``.
        normalization_layer (nn.Module, optional): Normalization layer.
        pretrained (bool, optional): Whether to load pretrained weights.
            Defaults to ``False``.

    Returns:
        _PackedResNet: A Packed-Ensembles ResNet.
    """
    block = _BasicBlock if arch in [18, 20, 34, 44, 56, 110, 1202] else _Bottleneck
    in_planes = 16 if arch in [20, 44, 56, 110, 1202] else 64
    net = _PackedResNet(
        block=block,
        num_blocks=get_resnet_num_blocks(arch),
        in_channels=in_channels,
        num_estimators=num_estimators,
        alpha=alpha,
        gamma=gamma,
        conv_bias=conv_bias,
        dropout_rate=dropout_rate,
        groups=groups,
        num_classes=num_classes,
        style=style,
        in_planes=int(in_planes * width_multiplier),
        normalization_layer=normalization_layer,
    )
    if pretrained:  # coverage: ignore
        weights = weight_ids[str(num_classes)][str(arch)]
        if weights is None:
            raise ValueError("No pretrained weights for this configuration")
        state_dict, config = load_hf(weights)
        if not net.check_config(config):
            raise ValueError("Pretrained weights do not match current configuration.")
        net.load_state_dict(state_dict)
    return net



# ruff: noqa: F401
from .wrappers import (
    EMA,
    EPOCH_UPDATE_MODEL,
    STEP_UPDATE_MODEL,
    SWA,
    SWAG,
    BatchEnsemble,
    CheckpointEnsemble,
    MCDropout,
    StochasticModel,
    batch_ensemble,
    deep_ensembles,
    mc_dropout,
)



import torch
from torch import Tensor, nn

from torch_uncertainty.layers.distributions import get_dist_conv_layer, get_dist_linear_layer

__all__ = [
    "dummy_model",
]


class _Dummy(nn.Module):
    def __init__(
        self,
        in_channels: int,
        num_classes: int,
        dropout_rate: float,
        dist_family: str | None = None,
    ) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.dropout_rate = dropout_rate

        self.linear = nn.Linear(1, num_classes)

        if dist_family is None:
            self.last_layer = nn.Linear(num_classes, num_classes)
        else:
            self.last_layer = get_dist_linear_layer(dist_family)(
                base_layer=nn.Linear,
                event_dim=num_classes,
                in_features=num_classes,
            )
        self.dropout = nn.Dropout(p=dropout_rate)

    def forward(self, x: Tensor) -> Tensor:
        return self.last_layer(
            self.dropout(
                self.linear(
                    torch.ones(
                        (x.shape[0], 1),
                        dtype=torch.float32,
                    )
                )
            )
        )


class _DummyWithFeats(_Dummy):
    def feats_forward(self, x: Tensor) -> Tensor:
        return torch.ones(
            (x.shape[0], 1),
            dtype=torch.float32,
        )


class _DummySegmentation(nn.Module):
    def __init__(
        self,
        in_channels: int,
        num_classes: int,
        dropout_rate: float,
        image_size: int,
        dist_family: str | None = None,
    ) -> None:
        super().__init__()
        self.dropout_rate = dropout_rate
        self.in_channels = in_channels
        self.num_classes = num_classes
        self.image_size = image_size
        self.conv = nn.Conv2d(in_channels, num_classes, kernel_size=3, padding=1)
        self.dropout = nn.Dropout(p=dropout_rate)
        if dist_family is None:
            self.last_layer = nn.Identity()
        else:
            self.last_layer = get_dist_conv_layer(dist_family)(
                base_layer=nn.Conv2d,
                event_dim=num_classes,
                in_channels=num_classes,
                kernel_size=3,
                padding=1,
            )

    def forward(self, x: Tensor) -> Tensor:
        return self.last_layer(
            self.dropout(
                self.conv(
                    torch.ones(
                        (
                            x.shape[0],
                            self.in_channels,
                            self.image_size,
                            self.image_size,
                        ),
                        dtype=torch.float32,
                    )
                )
            )
        )


def dummy_model(
    in_channels: int,
    num_classes: int,
    dropout_rate: float = 0.0,
    with_feats: bool = True,
    dist_family: str | None = None,
) -> _Dummy:
    """Dummy model for testing purposes.

    Args:
        in_channels (int): Number of input channels.
        num_classes (int): Number of output classes.
        num_estimators (int): Number of estimators in the ensemble.
        dropout_rate (float, optional): Dropout rate. Defaults to 0.0.
        with_feats (bool, optional): Whether to include features. Defaults to True.
        dist_family (str, optional): Distribution family. Defaults to None.

    Returns:
        _Dummy: Dummy model.
    """
    if with_feats:
        return _DummyWithFeats(
            in_channels=in_channels,
            num_classes=num_classes,
            dropout_rate=dropout_rate,
            dist_family=dist_family,
        )
    return _Dummy(
        in_channels=in_channels,
        num_classes=num_classes,
        dropout_rate=dropout_rate,
        dist_family=dist_family,
    )


def dummy_segmentation_model(
    in_channels: int,
    num_classes: int,
    image_size: int,
    dropout_rate: float = 0.0,
    dist_family: str | None = None,
) -> nn.Module:
    """Dummy segmentation model for testing purposes.

    Args:
        in_channels (int): Number of input channels.
        num_classes (int): Number of output classes.
        image_size (int): Size of the input image.
        dropout_rate (float, optional): Dropout rate. Defaults to 0.0.
        dist_family (str, optional): Distribution family. Defaults to None.

    Returns:
        nn.Module: Dummy segmentation model.
    """
    return _DummySegmentation(
        in_channels=in_channels,
        num_classes=num_classes,
        dropout_rate=dropout_rate,
        image_size=image_size,
        dist_family=dist_family,
    )



import numpy as np
import pytest
import torch
from PIL import Image
from torchvision import tv_tensors

from torch_uncertainty.transforms import (
    AutoContrast,
    Brightness,
    Color,
    Contrast,
    Equalize,
    MIMOBatchFormat,
    Posterize,
    RandomRescale,
    RepeatTarget,
    Rotate,
    Sharpen,
    Shear,
    Solarize,
    Translate,
)


@pytest.fixture()
def img_input() -> torch.Tensor:
    rng = np.random.default_rng()
    imarray = rng.uniform(low=0, high=255, size=(28, 28, 3))
    return Image.fromarray(imarray.astype("uint8")).convert("RGB")


@pytest.fixture()
def tv_tensors_input() -> tuple[torch.Tensor, torch.Tensor]:
    rng = np.random.default_rng()
    imarray1 = rng.uniform(low=0, high=255, size=(3, 28, 28))
    imarray2 = rng.uniform(low=0, high=255, size=(1, 28, 28))
    return (
        tv_tensors.Image(imarray1.astype("uint8")),
        tv_tensors.Mask(imarray2.astype("uint8")),
    )


@pytest.fixture()
def batch_input() -> tuple[torch.Tensor, torch.Tensor]:
    imgs = torch.rand(2, 3, 28, 28)
    return imgs, torch.tensor([0, 1])


class TestAutoContrast:
    """Testing the AutoContrast transform."""

    def test_pil(self, img_input):
        aug = AutoContrast()
        _ = aug(img_input)


class TestEqualize:
    """Testing the Equalize transform."""

    def test_pil(self, img_input):
        aug = Equalize()
        _ = aug(img_input)


class TestPosterize:
    """Testing the Posterize transform."""

    def test_pil(self, img_input):
        aug = Posterize()
        _ = aug(img_input, 2)

    def test_failures(self, img_input):
        aug = Posterize()
        with pytest.raises(ValueError):
            _ = aug(img_input, 5)
        with pytest.raises(ValueError):
            _ = aug(img_input, -1)


class TestSolarize:
    """Testing the Solarize transform."""

    def test_pil(self, img_input):
        aug = Solarize()
        _ = aug(img_input, 2)

    def test_failures(self, img_input):
        aug = Solarize()
        with pytest.raises(ValueError):
            _ = aug(img_input, 300)
        with pytest.raises(ValueError):
            _ = aug(img_input, -1)


class TestRotate:
    """Testing the Rotate transform."""

    def test_pil(self, img_input):
        aug = Rotate(random_direction=True)
        _ = aug(img_input, 10)


class TestShear:
    """Testing the Shear transform."""

    def test_pil(self, img_input):
        aug = Shear(axis=0, random_direction=True)
        _ = aug(img_input, 10)

    def test_failures(self):
        with pytest.raises(ValueError):
            _ = Shear(axis=2, random_direction=True)


class TestTranslate:
    """Testing the Translate transform."""

    def test_pil(self, img_input):
        aug = Translate(axis=0, random_direction=True)
        _ = aug(img_input, 10)

    def test_failures(self):
        with pytest.raises(ValueError):
            _ = Translate(axis=2, random_direction=True)


class TestContrast:
    """Testing the Contrast transform."""

    def test_pil(self, img_input):
        aug = Contrast()
        _ = aug(img_input, 2)

    def test_failures(self, img_input):
        aug = Contrast()
        with pytest.raises(ValueError):
            _ = aug(img_input, -1)


class TestBrightness:
    """Testing the Brightness transform."""

    def test_pil(self, img_input):
        aug = Brightness()
        _ = aug(img_input, 2)

    def test_failures(self, img_input):
        aug = Brightness()
        with pytest.raises(ValueError):
            _ = aug(img_input, -1)


class TestSharpen:
    """Testing the Sharpen transform."""

    def test_pil(self, img_input):
        aug = Sharpen()
        _ = aug(img_input, 2)

    def test_failures(self, img_input):
        aug = Sharpen()
        with pytest.raises(ValueError):
            _ = aug(img_input, -1)


class TestColor:
    """Testing the Color transform."""

    def test_pil(self, img_input):
        aug = Color()
        aug(img_input, 2)

    def test_tensor(self):
        aug = Color()
        aug(torch.rand(3, 28, 28), 2)

    def test_failures(self, img_input):
        aug = Color()
        with pytest.raises(ValueError):
            _ = aug(img_input, -1)


class TestRepeatTarget:
    """Testing the RepeatTarget transform."""

    def test_batch(self, batch_input):
        fn = RepeatTarget(3)
        _, target = fn(batch_input)
        assert target.shape == (6,)

    def test_failures(self):
        with pytest.raises(TypeError):
            _ = RepeatTarget(1.2)

        with pytest.raises(ValueError):
            _ = RepeatTarget(0)


class TestMIMOBatchFormat:
    """Testing the MIMOBatchFormat transform."""

    def test_batch(self, batch_input):
        b, c, h, w = batch_input[0].shape

        fn = MIMOBatchFormat(1, 0, 1)
        imgs, target = fn(batch_input)
        assert imgs.shape == (b, c, h, w)
        assert target.shape == (b,)

        fn = MIMOBatchFormat(4, 0, 2)
        imgs, target = fn(batch_input)
        assert imgs.shape == (b * 4 * 2, 3, 28, 28)
        assert target.shape == (b * 4 * 2,)

    def test_failures(self):
        with pytest.raises(ValueError):
            _ = MIMOBatchFormat(0, 0, 1)

        with pytest.raises(ValueError):
            _ = MIMOBatchFormat(1, -1, 1)

        with pytest.raises(ValueError):
            _ = MIMOBatchFormat(1, 1.2, 1)

        with pytest.raises(ValueError):
            _ = MIMOBatchFormat(1, 0, 0)


class TestRandomRescale:
    """Testing the RandomRescale transform."""

    def test_tv_tensors(self, tv_tensors_input):
        aug = RandomRescale(0.5, 2.0)
        _ = aug(tv_tensors_input)



import sys

from torch_uncertainty.baselines.classification import ResNetBaseline
from torch_uncertainty.datamodules import CIFAR10DataModule
from torch_uncertainty.utils.cli import TULightningCLI, TUSaveConfigCallback


class TestCLI:
    """Testing torch-uncertainty CLI."""

    def test_cli_init(self):
        """Test CLI initialization."""
        sys.argv = [
            "file.py",
            "--model.in_channels",
            "3",
            "--model.num_classes",
            "10",
            "--model.version",
            "std",
            "--model.arch",
            "18",
            "--model.loss.class_path",
            "torch.nn.CrossEntropyLoss",
            "--data.root",
            "./data",
            "--data.batch_size",
            "4",
            "--trainer.callbacks+=ModelCheckpoint",
            "--trainer.callbacks.monitor=val/cls/Acc",
            "--trainer.callbacks.mode=max",
        ]
        cli = TULightningCLI(ResNetBaseline, CIFAR10DataModule, run=False)
        assert cli.eval_after_fit_default is False
        assert cli.save_config_callback == TUSaveConfigCallback
        assert isinstance(cli.trainer.callbacks[0], TUSaveConfigCallback)
        cli.trainer.callbacks[0].setup(cli.trainer, cli.model, stage="fit")
        cli.trainer.callbacks[0].already_saved = True
        cli.trainer.callbacks[0].setup(cli.trainer, cli.model, stage="fit")



import logging
from collections.abc import Callable
from pathlib import Path

from torchvision.datasets import ImageFolder
from torchvision.datasets.utils import (
    check_integrity,
    download_and_extract_archive,
)


class OpenImageO(ImageFolder):
    url = "https://zenodo.org/records/10540831/files/OpenImage-O.zip"
    filename = "OpenImage-O.zip"
    md5sum = "c0abd7cd4b6f218a7149adc718d70e6e"

    def __init__(
        self,
        root: str | Path,
        split: str | None = None,
        transform: Callable | None = None,
        target_transform: Callable | None = None,
        download: bool = False,
    ) -> None:
        """OpenImage-O dataset.

        Args:
            root (str): Root directory of the datasets.
            split (str, optional): Unused, for API consistency. Defaults to
                None.
            transform (callable, optional): A function/transform that takes in
                a PIL image and returns a transformed version. E.g,
                ``transforms.RandomCrop``. Defaults to None.
            target_transform (callable, optional): A function/transform that
                takes in the target and transforms it. Defaults to None.
            download (bool, optional): If True, downloads the dataset from the
                internet and puts it in root directory. If dataset is already
                downloaded, it is not downloaded again. Defaults to False.

        References:
            Original dataset: The open images dataset v4: Unified image
            classification, object detection, and visual relationship detection
            at scale. Kuznetsova, A., et al. The International Journal of
            Computer Vision.

            Curation: ViM: Out-Of-Distribution with Virtual-logit Matching.
            Wang H., et al. In CVPR 2022.
        """
        self.root = Path(root)
        self.split = split

        self.transform = transform
        self.target_transform = target_transform

        if download:
            self.download()

        super().__init__(
            self.root / "openimage-o/",
            transform=transform,
            target_transform=target_transform,
        )

    def _check_integrity(self) -> bool:
        fpath = self.root / self.filename
        return check_integrity(
            fpath,
            self.md5sum,
        )

    def download(self) -> None:
        if self._check_integrity():
            logging.info("Files already downloaded and verified")
            return

        download_and_extract_archive(
            self.url,
            download_root=self.root,
            extract_root=self.root / "openimage-o/ood/",
            filename=self.filename,
            md5=self.md5sum,
        )



import pytest

from tests._dummies.dataset import DummySegmentationDataset
from torch_uncertainty.datamodules.segmentation import MUADDataModule
from torch_uncertainty.datasets import MUAD


class TestMUADDataModule:
    """Testing the MUADDataModule datamodule."""

    def test_camvid_main(self):
        dm = MUADDataModule(root="./data/", batch_size=128)

        assert dm.dataset == MUAD

        dm.dataset = DummySegmentationDataset

        dm.prepare_data()
        dm.setup()

        with pytest.raises(ValueError):
            dm.setup("xxx")

        # test abstract methods
        dm.get_train_set()
        dm.get_val_set()
        dm.get_test_set()

        dm.train_dataloader()
        dm.val_dataloader()
        dm.test_dataloader()

        dm.val_split = 0.1
        dm.prepare_data()
        dm.setup()
        dm.train_dataloader()
        dm.val_dataloader()

    def test_small_muad_accessibility(self):
        dataset = MUAD(root="./data/", split="test", version="small", download=True)
        assert len(dataset.samples) > 0, "Dataset is not found"



import pytest
import torch
from torch import nn

from torch_uncertainty.baselines.classification import (
    ResNetBaseline,
    VGGBaseline,
    WideResNetBaseline,
)
from torch_uncertainty.baselines.regression import MLPBaseline


class TestPackedBaseline:
    """Testing the PackedResNet baseline class."""

    def test_packed_50(self):
        net = ResNetBaseline(
            num_classes=10,
            in_channels=3,
            loss=nn.CrossEntropyLoss(),
            version="packed",
            arch=50,
            style="cifar",
            num_estimators=4,
            alpha=2,
            gamma=1,
            groups=1,
        )

        _ = net(torch.rand(1, 3, 32, 32))

    def test_packed_18(self):
        net = ResNetBaseline(
            num_classes=10,
            in_channels=3,
            loss=nn.CrossEntropyLoss(),
            version="packed",
            arch=18,
            style="imagenet",
            num_estimators=4,
            alpha=2,
            gamma=2,
            groups=2,
        )

        _ = net(torch.rand(1, 3, 40, 40))

    def test_packed_exception(self):
        with pytest.raises(ValueError):
            _ = ResNetBaseline(
                num_classes=10,
                in_channels=3,
                loss=nn.CrossEntropyLoss(),
                version="packed",
                arch=50,
                style="cifar",
                num_estimators=4,
                alpha=0,
                gamma=1,
                groups=1,
            )

        with pytest.raises(ValueError):
            _ = ResNetBaseline(
                num_classes=10,
                in_channels=3,
                loss=nn.CrossEntropyLoss(),
                version="packed",
                arch=50,
                style="cifar",
                num_estimators=4,
                alpha=2,
                gamma=0,
                groups=1,
            )


class TestPackedWideBaseline:
    """Testing the PackedWideResNet baseline class."""

    def test_packed(self):
        net = WideResNetBaseline(
            num_classes=10,
            in_channels=3,
            loss=nn.CrossEntropyLoss(),
            version="packed",
            style="cifar",
            num_estimators=4,
            alpha=2,
            gamma=1,
            groups=1,
        )

        _ = net(torch.rand(1, 3, 32, 32))


class TestPackedVGGBaseline:
    """Testing the PackedWideResNet baseline class."""

    def test_packed(self):
        net = VGGBaseline(
            num_classes=10,
            in_channels=3,
            arch=13,
            loss=nn.CrossEntropyLoss(),
            version="packed",
            num_estimators=4,
            alpha=2,
            gamma=1,
            groups=1,
        )
        _ = net(torch.rand(2, 3, 32, 32))


class TestPackedMLPBaseline:
    """Testing the Packed MLP baseline class."""

    def test_packed(self):
        net = MLPBaseline(
            in_features=3,
            output_dim=10,
            loss=nn.MSELoss(),
            version="packed",
            hidden_dims=[1],
            num_estimators=2,
            alpha=2,
            gamma=1,
        )
        _ = net(torch.rand(1, 3))



import logging
from collections.abc import Callable
from pathlib import Path

import numpy as np
from torch import Tensor
from torchvision.datasets import VisionDataset
from torchvision.datasets.utils import (
    check_integrity,
    download_and_extract_archive,
)


class CIFAR10C(VisionDataset):
    """The corrupted CIFAR-10-C Dataset.

    Args:
        root (str): Root directory of the datasets.
        transform (callable, optional): A function/transform that takes in
            a PIL image and returns a transformed version. E.g,
            ``transforms.RandomCrop``. Defaults to None.
        target_transform (callable, optional): A function/transform that
            takes in the target and transforms it. Defaults to None.
        subset (str): The subset to use, one of ``all`` or the keys in
            ``cifarc_subsets``.
        shift_severity (int): The shift_severity of the corruption, between 1 and 5.
        download (bool, optional): If True, downloads the dataset from the
            internet and puts it in root directory. If dataset is already
            downloaded, it is not downloaded again. Defaults to False.

    References:
        Benchmarking neural network robustness to common corruptions and
        perturbations. Dan Hendrycks and Thomas Dietterich. In ICLR, 2019.
    """

    base_folder = "CIFAR-10-C"
    tgz_md5 = "56bf5dcef84df0e2308c6dcbcbbd8499"
    cifarc_subsets = [
        "brightness",
        "contrast",
        "defocus_blur",
        "elastic_transform",
        "fog",
        "frost",
        "gaussian_blur",
        "gaussian_noise",
        "glass_blur",
        "impulse_noise",
        "jpeg_compression",
        "motion_blur",
        "pixelate",
        "saturate",
        "shot_noise",
        "snow",
        "spatter",
        "speckle_noise",
        "zoom_blur",
    ]

    ctest_list = [
        ["fog.npy", "7b397314b5670f825465fbcd1f6e9ccd"],
        ["jpeg_compression.npy", "2b9cc4c864e0193bb64db8d7728f8187"],
        ["zoom_blur.npy", "6ea8e63f1c5cdee1517533840641641b"],
        ["speckle_noise.npy", "ef00b87611792b00df09c0b0237a1e30"],
        ["glass_blur.npy", "7361fb4019269e02dbf6925f083e8629"],
        ["spatter.npy", "8a5a3903a7f8f65b59501a6093b4311e"],
        ["shot_noise.npy", "3a7239bb118894f013d9bf1984be7f11"],
        ["defocus_blur.npy", "7d1322666342a0702b1957e92f6254bc"],
        ["elastic_transform.npy", "9421657c6cd452429cf6ce96cc412b5f"],
        ["gaussian_blur.npy", "c33370155bc9b055fb4a89113d3c559d"],
        ["frost.npy", "31f6ab3bce1d9934abfb0cc13656f141"],
        ["saturate.npy", "1cfae0964219c5102abbb883e538cc56"],
        ["brightness.npy", "0a81ef75e0b523c3383219c330a85d48"],
        ["snow.npy", "bb238de8555123da9c282dea23bd6e55"],
        ["gaussian_noise.npy", "ecaf8b9a2399ffeda7680934c33405fd"],
        ["motion_blur.npy", "fffa5f852ff7ad299cfe8a7643f090f4"],
        ["contrast.npy", "3c8262171c51307f916c30a3308235a8"],
        ["impulse_noise.npy", "2090e01c83519ec51427e65116af6b1a"],
        ["labels.npy", "c439b113295ed5254878798ffe28fd54"],
        ["pixelate.npy", "0f14f7e2db14288304e1de10df16832f"],
    ]
    url = "https://zenodo.org/record/2535967/files/CIFAR-10-C.tar"
    filename = "CIFAR-10-C.tar"

    def __init__(
        self,
        root: Path | str,
        transform: Callable | None = None,
        target_transform: Callable | None = None,
        subset: str = "all",
        shift_severity: int = 1,
        download: bool = False,
    ) -> None:
        self.root = Path(root)
        # Download the new targets
        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError("Dataset not found. You can use download=True to download it.")

        super().__init__(
            root=self.root / self.base_folder,
            transform=transform,
            target_transform=target_transform,
        )
        if subset not in ["all", *self.cifarc_subsets]:
            raise ValueError(f"The subset '{subset}' does not exist in CIFAR-C.")
        self.subset = subset
        self.shift_severity = shift_severity

        if shift_severity not in list(range(1, 6)):
            raise ValueError(
                "Corruptions shift_severity should be chosen between 1 and 5 included."
            )
        samples, labels = self.make_dataset(self.root, self.subset, self.shift_severity)

        self.samples = samples
        self.labels = labels.astype(np.int64)

    def make_dataset(
        self, root: Path, subset: str, shift_severity: int
    ) -> tuple[np.ndarray, np.ndarray]:
        r"""Make the CIFAR-C dataset.

        Build the corrupted dataset according to the chosen subset and
            shift_severity. If the subset is 'all', gather all corruption types
            in the dataset.

        Args:
            root (Path):The path to the dataset.
            subset (str): The name of the corruption subset to be used. Choose
                `all` for the dataset to contain all subsets.
            shift_severity (int): The shift_severity of the corruption applied to the
                images.

        Returns:
            Tuple[np.ndarray, np.ndarray]: The samples and labels of the chosen.
        """
        if subset == "all":
            labels: np.ndarray = np.load(root / "labels.npy")[
                (shift_severity - 1) * 10000 : shift_severity * 10000
            ]
            sample_arrays = [
                np.load(root / (cifar_subset + ".npy"))[
                    (shift_severity - 1) * 10000 : shift_severity * 10000
                ]
                for cifar_subset in self.cifarc_subsets
            ]
            samples = np.concatenate(sample_arrays, axis=0)
            labels = np.tile(labels, len(self.cifarc_subsets))

        else:
            samples: np.ndarray = np.load(root / (subset + ".npy"))[
                (shift_severity - 1) * 10000 : shift_severity * 10000
            ]
            labels: np.ndarray = np.load(root / "labels.npy")[
                (shift_severity - 1) * 10000 : shift_severity * 10000
            ]
        return samples, labels

    def __len__(self) -> int:
        """The number of samples in the dataset."""
        return self.labels.shape[0]

    def __getitem__(self, index: int) -> tuple[np.ndarray | Tensor, int]:
        """Get the samples and targets of the dataset.

        Args:
            index (int): The index of the sample to get.
        """
        sample, target = (
            self.samples[index],
            self.labels[index],
        )

        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return sample, target

    def _check_integrity(self) -> bool:
        """Check the integrity of the dataset."""
        for filename, md5 in self.ctest_list:
            fpath = self.root / self.base_folder / filename
            if not check_integrity(fpath, md5):
                return False
        return True

    def download(self) -> None:
        """Download the dataset."""
        if self._check_integrity():
            logging.info("Files already downloaded and verified")
            return
        download_and_extract_archive(self.url, self.root, filename=self.filename, md5=self.tgz_md5)


class CIFAR100C(CIFAR10C):
    base_folder = "CIFAR-100-C"
    tgz_md5 = "11f0ed0f1191edbf9fa23466ae6021d3"
    ctest_list = [
        ["fog.npy", "4efc7ebd5e82b028bdbe13048e3ea564"],
        ["jpeg_compression.npy", "c851b7f1324e1d2ffddeb76920576d11"],
        ["zoom_blur.npy", "0204613400c034a81c4830d5df81cb82"],
        ["speckle_noise.npy", "e3f215b1a0f9fd9fd6f0d1cf94a7ce99"],
        ["glass_blur.npy", "0bf384f38e5ccbf8dd479d9059b913e1"],
        ["spatter.npy", "12ccf41d62564d36e1f6a6ada5022728"],
        ["shot_noise.npy", "b0a1fa6e1e465a747c1b204b1914048a"],
        ["defocus_blur.npy", "d923e3d9c585a27f0956e2f2ad832564"],
        ["elastic_transform.npy", "a0792bd6581f6810878be71acedfc65a"],
        ["gaussian_blur.npy", "5204ba0d557839772ef5a4196a052c3e"],
        ["frost.npy", "3a39c6823bdfaa0bf8b12fe7004b8117"],
        ["saturate.npy", "c0697e9fdd646916a61e9c312c77bf6b"],
        ["brightness.npy", "f22d7195aecd6abb541e27fca230c171"],
        ["snow.npy", "0237be164583af146b7b144e73b43465"],
        ["gaussian_noise.npy", "ecc4d366eac432bdf25c024086f5e97d"],
        ["motion_blur.npy", "732a7e2e54152ff97c742d4c388c5516"],
        ["contrast.npy", "322bb385f1d05154ee197ca16535f71e"],
        ["impulse_noise.npy", "3b3c210ddfa0b5cb918ff4537a429fef"],
        ["labels.npy", "bb4026e9ce52996b95f439544568cdb2"],
        ["pixelate.npy", "96c00c60f144539e14cffb02ddbd0640"],
    ]
    url = "https://zenodo.org/record/3555552/files/CIFAR-100-C.tar"
    filename = "CIFAR-100-C.tar"



import torch
from torch import Tensor
from torch.nn import Module, init
from torch.nn import functional as F
from torch.nn.common_types import _size_1_t, _size_2_t, _size_3_t
from torch.nn.modules.utils import (
    _pair,
    _reverse_repeat_tuple,
    _single,
    _triple,
)
from torch.nn.parameter import Parameter

from .sampler import CenteredGaussianMixture, TrainableDistribution

__all__ = ["BayesConv1d", "BayesConv2d", "BayesConv3d"]


class _BayesConvNd(Module):
    __constants__ = [
        "stride",
        "padding",
        "dilation",
        "groups",
        "padding_mode",
        "output_padding",
        "in_channels",
        "out_channels",
        "kernel_size",
    ]
    __annotations__ = {"bias": torch.Tensor | None}

    def _conv_forward(
        self, inputs: Tensor, weight: Tensor, bias: Tensor | None
    ) -> Tensor:  # coverage: ignore
        ...

    in_channels: int
    _reversed_padding_repeated_twice: list[int]
    out_channels: int
    kernel_size: tuple[int, ...]
    stride: tuple[int, ...]
    padding: str | tuple[int, ...]
    dilation: tuple[int, ...]
    prior_mu: float
    prior_sigma: float
    mu_init: float
    sigma_init: float
    frozen: bool
    transposed: bool
    output_padding: tuple[int, ...]
    groups: int
    padding_mode: str
    weight: Tensor
    bias: Tensor | None
    lprior: Tensor
    lvposterior: Tensor

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, ...],
        stride: tuple[int, ...],
        padding: tuple[int, ...],
        dilation: tuple[int, ...],
        prior_sigma_1: float,
        prior_sigma_2: float,
        prior_pi: float,
        mu_init: float,
        sigma_init: float,
        frozen: bool,
        transposed: bool,
        output_padding: tuple[int, ...],
        groups: int,
        bias: bool,
        padding_mode: str,
        device=None,
        dtype=None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()

        valid_padding_modes = {"zeros", "reflect", "replicate", "circular"}
        if padding_mode not in valid_padding_modes:
            raise ValueError(
                f"padding_mode must be one of {valid_padding_modes}, but got '{padding_mode}'"
            )

        if transposed:
            raise NotImplementedError(
                "Bayesian transposed convolution not implemented yet. Raise an issue if needed."
            )

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.prior_sigma_1 = prior_sigma_1
        self.prior_sigma_2 = prior_sigma_2
        self.prior_pi = prior_pi
        self.mu_init = mu_init
        self.sigma_init = sigma_init
        self.frozen = frozen
        self.output_padding = output_padding
        self.groups = groups
        self.padding_mode = padding_mode

        self._reversed_padding_repeated_twice = _reverse_repeat_tuple(self.padding, 2)

        self.weight_mu = Parameter(
            torch.empty(
                (out_channels, in_channels // groups, *kernel_size),
                **factory_kwargs,
            )
        )
        self.weight_sigma = Parameter(
            torch.empty(
                (out_channels, in_channels // groups, *kernel_size),
                **factory_kwargs,
            )
        )

        if bias:
            self.bias_mu = Parameter(torch.empty(out_channels, **factory_kwargs))
            self.bias_sigma = Parameter(torch.empty(out_channels, **factory_kwargs))
        else:
            self.register_parameter("bias_mu", None)
            self.register_parameter("bias_sigma", None)

        self.weight_prior_dist = CenteredGaussianMixture(prior_sigma_1, prior_sigma_2, prior_pi)
        if bias:
            self.bias_prior_dist = CenteredGaussianMixture(prior_sigma_1, prior_sigma_2, prior_pi)

        self.reset_parameters()

        self.weight_sampler = TrainableDistribution(self.weight_mu, self.weight_sigma)
        if bias:
            self.bias_sampler = TrainableDistribution(self.bias_mu, self.bias_sigma)

    def reset_parameters(self) -> None:
        # TODO: change init
        init.normal_(self.weight_mu, mean=self.mu_init, std=0.1)
        init.normal_(self.weight_sigma, mean=self.sigma_init, std=0.1)

        if self.bias_mu is not None:
            init.normal_(self.bias_mu, mean=self.mu_init, std=0.1)
            init.normal_(self.bias_sigma, mean=self.sigma_init, std=0.1)

    def freeze(self) -> None:
        """Freeze the layer by setting the frozen attribute to True."""
        self.frozen = True

    def unfreeze(self) -> None:
        """Unfreeze the layer by setting the frozen attribute to False."""
        self.frozen = False

    def sample(self) -> tuple[Tensor, Tensor | None]:
        """Sample the Bayesian layer's posterior."""
        weight = self.weight_sampler.sample()
        bias = self.bias_sampler.sample() if self.bias_mu is not None else None
        return weight, bias

    def extra_repr(self) -> str:  # coverage: ignore
        s = "{in_channels}, {out_channels}, kernel_size={kernel_size}, stride={stride}"
        if self.padding != (0,) * len(self.padding):
            s += ", padding={padding}"
        if self.dilation != (1,) * len(self.dilation):
            s += ", dilation={dilation}"
        if self.output_padding != (0,) * len(self.output_padding):
            s += ", output_padding={output_padding}"
        if self.groups != 1:
            s += ", groups={groups}"
        if self.bias_mu is None:
            s += ", bias=False"
        if self.padding_mode != "zeros":
            s += ", padding_mode={padding_mode}"
        return s.format(**self.__dict__)

    def __setstate__(self, state) -> None:
        super().__setstate__(state)
        if not hasattr(self, "padding_mode"):  # coverage: ignore
            self.padding_mode = "zeros"


class BayesConv1d(_BayesConvNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_1_t,
        stride: _size_1_t = 1,
        padding: str | _size_1_t = 0,
        dilation: _size_1_t = 1,
        prior_sigma_1: float = 0.1,
        prior_sigma_2: float = 0.002,
        prior_pi: float = 1,
        mu_init: float = 0.0,
        sigma_init: float = -6.0,
        frozen: bool = False,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",  # TODO: refine this type
        device=None,
        dtype=None,
    ) -> None:
        """Bayesian Conv1d Layer with Mixture of Normals prior and Normal
        posterior.
        """
        factory_kwargs = {"device": device, "dtype": dtype}
        kernel_size_ = _single(kernel_size)
        stride_ = _single(stride)
        padding_ = padding if isinstance(padding, str) else _single(padding)
        dilation_ = _single(dilation)
        super().__init__(
            in_channels,
            out_channels,
            kernel_size_,
            stride_,
            padding_,
            dilation_,
            prior_sigma_1,
            prior_sigma_2,
            prior_pi,
            mu_init,
            sigma_init,
            frozen,
            False,
            _single(0),
            groups,
            bias,
            padding_mode,
            **factory_kwargs,
        )

    def _conv_forward(self, inputs: Tensor, weight: Tensor, bias: Tensor | None) -> Tensor:
        if self.padding_mode != "zeros":
            return F.conv1d(
                F.pad(
                    inputs,
                    self._reversed_padding_repeated_twice,
                    mode=self.padding_mode,
                ),
                weight,
                bias,
                self.stride,
                _single(0),
                self.dilation,
                self.groups,
            )
        return F.conv1d(
            inputs,
            weight,
            bias,
            self.stride,
            self.padding,
            self.dilation,
            self.groups,
        )

    def forward(self, inputs: Tensor) -> Tensor:
        if self.frozen:
            weight = self.weight_mu
            bias = self.bias_mu
        else:
            weight = self.weight_sampler.sample()

            if self.bias_mu is not None:
                bias = self.bias_sampler.sample()
                bias_lposterior = self.bias_sampler.log_posterior()
                bias_lprior = self.bias_prior_dist.log_prob(bias)
            else:
                bias, bias_lposterior, bias_lprior = None, 0, 0

            self.lvposterior = self.weight_sampler.log_posterior() + bias_lposterior
            self.lprior = self.weight_prior_dist.log_prob(weight) + bias_lprior

        return self._conv_forward(inputs, weight, bias)


class BayesConv2d(_BayesConvNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_2_t,
        stride: _size_2_t = 1,
        padding: str | _size_2_t = 0,
        dilation: _size_2_t = 1,
        prior_sigma_1: float = 0.1,
        prior_sigma_2: float = 0.002,
        prior_pi: float = 1,
        mu_init: float = 0.0,
        sigma_init: float = -6.0,
        frozen: bool = False,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",  # TODO: refine this type
        device=None,
        dtype=None,
    ) -> None:
        """Bayesian Conv2d Layer with Gaussian Mixture prior and Normal posterior."""
        factory_kwargs = {"device": device, "dtype": dtype}
        kernel_size_ = _pair(kernel_size)
        stride_ = _pair(stride)
        padding_ = padding if isinstance(padding, str) else _pair(padding)
        dilation_ = _pair(dilation)
        super().__init__(
            in_channels,
            out_channels,
            kernel_size_,
            stride_,
            padding_,
            dilation_,
            prior_sigma_1,
            prior_sigma_2,
            prior_pi,
            mu_init,
            sigma_init,
            frozen,
            False,
            _pair(0),
            groups,
            bias,
            padding_mode,
            **factory_kwargs,
        )

    def _conv_forward(self, inputs: Tensor, weight: Tensor, bias: Tensor | None) -> Tensor:
        if self.padding_mode != "zeros":
            return F.conv2d(
                F.pad(
                    inputs,
                    self._reversed_padding_repeated_twice,
                    mode=self.padding_mode,
                ),
                weight,
                bias,
                self.stride,
                _pair(0),
                self.dilation,
                self.groups,
            )
        return F.conv2d(
            inputs,
            weight,
            bias,
            self.stride,
            self.padding,
            self.dilation,
            self.groups,
        )

    def forward(self, inputs: Tensor) -> Tensor:
        if self.frozen:
            weight = self.weight_mu
            bias = self.bias_mu
        else:
            weight = self.weight_sampler.sample()

            if self.bias_mu is not None:
                bias = self.bias_sampler.sample()
                bias_lposterior = self.bias_sampler.log_posterior()
                bias_lprior = self.bias_prior_dist.log_prob(bias)
            else:
                bias, bias_lposterior, bias_lprior = None, 0, 0

            self.lvposterior = self.weight_sampler.log_posterior() + bias_lposterior
            self.lprior = self.weight_prior_dist.log_prob(weight) + bias_lprior

        return self._conv_forward(inputs, weight, bias)


class BayesConv3d(_BayesConvNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_3_t,
        stride: _size_3_t = 1,
        padding: str | _size_3_t = 0,
        dilation: _size_3_t = 1,
        prior_sigma_1: float = 0.1,
        prior_sigma_2: float = 0.002,
        prior_pi: float = 1,
        mu_init: float = 0.0,
        sigma_init: float = 10.0,
        frozen: bool = False,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
        device=None,
        dtype=None,
    ) -> None:
        """Bayesian Conv3d Layer with Gaussian mixture prior and Normal posterior."""
        factory_kwargs = {"device": device, "dtype": dtype}
        kernel_size_ = _triple(kernel_size)
        stride_ = _triple(stride)
        padding_ = padding if isinstance(padding, str) else _triple(padding)
        dilation_ = _triple(dilation)
        super().__init__(
            in_channels,
            out_channels,
            kernel_size_,
            stride_,
            padding_,
            dilation_,
            prior_sigma_1,
            prior_sigma_2,
            prior_pi,
            mu_init,
            sigma_init,
            frozen,
            False,
            _triple(0),
            groups,
            bias,
            padding_mode,
            **factory_kwargs,
        )

    def _conv_forward(self, inputs: Tensor, weight: Tensor, bias: Tensor | None) -> Tensor:
        if self.padding_mode != "zeros":
            return F.conv3d(
                F.pad(
                    inputs,
                    self._reversed_padding_repeated_twice,
                    mode=self.padding_mode,
                ),
                weight,
                bias,
                self.stride,
                _triple(0),
                self.dilation,
                self.groups,
            )
        return F.conv3d(
            inputs,
            weight,
            bias,
            self.stride,
            self.padding,
            self.dilation,
            self.groups,
        )

    def forward(self, inputs: Tensor) -> Tensor:
        if self.frozen:
            weight = self.weight_mu
            bias = self.bias_mu
        else:
            weight = self.weight_sampler.sample()

            if self.bias_mu is not None:
                bias = self.bias_sampler.sample()
                bias_lposterior = self.bias_sampler.log_posterior()
                bias_lprior = self.bias_prior_dist.log_prob(bias)
            else:
                bias, bias_lposterior, bias_lprior = None, 0, 0

            self.lvposterior = self.weight_sampler.log_posterior() + bias_lposterior
            self.lprior = self.weight_prior_dist.log_prob(weight) + bias_lprior

        return self._conv_forward(inputs, weight, bias)



import logging
from collections.abc import Callable
from functools import partial
from typing import Literal

import torch
from timm.optim import Lamb
from torch import nn, optim
from torch.optim import Optimizer
from torch.optim.lr_scheduler import (
    CosineAnnealingLR,
    LinearLR,
    LRScheduler,
    MultiStepLR,
    SequentialLR,
)


def optim_abnn(
    model: nn.Module,
    lr: float,
    momentum: float = 0.9,
    weight_decay: float = 1e-4,
    nesterov: bool = True,
) -> dict:
    """ABNN finetuning recipe."""
    optimizer = optim.SGD(
        model.parameters(),
        lr=lr,
        momentum=momentum,
        weight_decay=weight_decay,
        nesterov=nesterov,
    )
    scheduler = MultiStepLR(
        optimizer,
        milestones=[1, 4],
        gamma=0.1,
    )
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def optim_cifar10_resnet18(
    model: nn.Module,
) -> dict[str, Optimizer | LRScheduler]:
    """Optimizer to train a ResNet18 on CIFAR-10."""
    optimizer = optim.SGD(
        model.parameters(),
        lr=0.1,
        momentum=0.9,
        weight_decay=5e-4,
        nesterov=True,
    )
    scheduler = MultiStepLR(
        optimizer,
        milestones=[25, 50],
        gamma=0.1,
    )
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def optim_cifar10_resnet50(
    model: nn.Module,
) -> dict[str, Optimizer | LRScheduler]:
    r"""Hyperparameters from Deep Residual Learning for Image Recognition
    https://arxiv.org/pdf/1512.03385.pdf.
    """
    optimizer = optim.SGD(
        model.parameters(),
        lr=0.1,
        momentum=0.9,
        weight_decay=5e-4,
        nesterov=True,
    )
    scheduler = MultiStepLR(
        optimizer,
        milestones=[60, 120, 160],
        gamma=0.2,
    )
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def optim_cifar10_wideresnet(
    model: nn.Module,
) -> dict[str, Optimizer | LRScheduler]:
    """Optimizer to train a WideResNet28x10 on CIFAR-10."""
    optimizer = optim.SGD(
        model.parameters(),
        lr=0.1,
        momentum=0.9,
        weight_decay=5e-4,
        nesterov=True,
    )
    scheduler = MultiStepLR(
        optimizer,
        milestones=[60, 120, 160],
        gamma=0.2,
    )
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def optim_cifar10_vgg16(
    model: nn.Module,
) -> dict[str, Optimizer | LRScheduler]:
    """Optimizer to train a VGG16 on CIFAR-10."""
    optimizer = optim.Adam(
        model.parameters(),
        lr=0.005,
        weight_decay=1e-6,
    )
    scheduler = MultiStepLR(
        optimizer,
        milestones=[25, 50],
        gamma=0.1,
    )
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def optim_cifar100_resnet18(
    model: nn.Module,
) -> dict[str, Optimizer | LRScheduler]:
    optimizer = optim.SGD(
        model.parameters(),
        lr=0.1,
        momentum=0.9,
        weight_decay=5e-4,
        nesterov=True,
    )
    scheduler = MultiStepLR(
        optimizer,
        milestones=[25, 50],
        gamma=0.1,
    )
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def optim_cifar100_resnet50(
    model: nn.Module,
) -> dict[str, Optimizer | LRScheduler]:
    r"""Hyperparameters from Deep Residual Learning for Image Recognition
    https://arxiv.org/pdf/1512.03385.pdf.
    """
    optimizer = optim.SGD(
        model.parameters(),
        lr=0.1,
        momentum=0.9,
        weight_decay=5e-4,
        nesterov=True,
    )
    scheduler = MultiStepLR(
        optimizer,
        milestones=[60, 120, 160],
        gamma=0.2,
    )
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def optim_cifar100_vgg16(
    model: nn.Module,
) -> dict[str, Optimizer | LRScheduler]:
    """Optimizer to train a VGG16 on CIFAR-100."""
    optimizer = optim.SGD(
        model.parameters(),
        lr=0.05,
        momentum=0.9,
        weight_decay=1e-4,
        nesterov=True,
    )
    scheduler = MultiStepLR(
        optimizer,
        milestones=[60, 120, 160],
        gamma=0.2,
    )
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def optim_imagenet_resnet50(
    model: nn.Module,
    num_epochs: int = 90,
    start_lr: float = 0.256,
    end_lr: float = 0,
) -> dict:
    r"""Hyperparameters from Deep Residual Learning for Image Recognition
    https://arxiv.org/pdf/1512.03385.pdf.
    """
    optimizer = optim.SGD(
        model.parameters(),
        lr=start_lr,
        momentum=0.875,
        weight_decay=3.0517578125e-05,
        nesterov=False,
    )
    scheduler = CosineAnnealingLR(optimizer, num_epochs, eta_min=end_lr)
    return {
        "optimizer": optimizer,
        "lr_scheduler": scheduler,
        "monitor": "cls/val_acc",
    }


def optim_imagenet_resnet50_a3(model: nn.Module, effective_batch_size: int | None = None) -> dict:
    """Training procedure proposed in ResNet strikes back: An improved training
        procedure in timm.

    Args:
        model (nn.Module): The model to be optimized.
        effective_batch_size (int, optional): The batch size of the model
            (taking multiple GPUs into account). Defaults to None.

    Returns:
        dict: The optimizer and the scheduler for the training.
    """
    if effective_batch_size is None:
        logging.warning("Setting effective batch size to 2048 for steps computations !")
        effective_batch_size = 2048

    optimizer = Lamb(model.parameters(), lr=0.008, weight_decay=0.02)

    warmup = LinearLR(
        optimizer,
        start_factor=1e-4,
        end_factor=1,
        total_iters=5 * (1281167 // effective_batch_size + 1),
    )
    cosine_scheduler = CosineAnnealingLR(
        optimizer,
        eta_min=1e-6,
        T_max=105 * (1281167 // effective_batch_size + 1),
    )
    scheduler = SequentialLR(
        optimizer,
        schedulers=[warmup, cosine_scheduler],
        milestones=[5 * (1281167 // effective_batch_size + 1)],
    )
    return {
        "optimizer": optimizer,
        "lr_scheduler": {
            "scheduler": scheduler,
            "interval": "step",
            "frequency": 1,
        },
        "monitor": "cls/val_acc",
    }


def optim_cifar10_resnet34(
    model: nn.Module,
) -> dict[str, Optimizer | LRScheduler]:
    optimizer = optim.SGD(
        model.parameters(),
        lr=0.1,
        momentum=0.9,
        weight_decay=1e-4,
        nesterov=True,
    )
    scheduler = MultiStepLR(
        optimizer,
        milestones=[100, 150],
        gamma=0.1,
    )
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def optim_cifar100_resnet34(
    model: nn.Module,
) -> dict[str, Optimizer | LRScheduler]:
    optimizer = optim.SGD(
        model.parameters(),
        lr=0.1,
        momentum=0.9,
        weight_decay=1e-4,
        nesterov=True,
    )
    scheduler = MultiStepLR(
        optimizer,
        milestones=[100, 150],
        gamma=0.1,
    )
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def optim_tinyimagenet_resnet34(
    model: nn.Module,
) -> dict[str, Optimizer | LRScheduler]:
    """Optimization recipe from 'The Devil is in the Margin: Margin-based
    Label Smoothing for Network Calibration',
    (CVPR 2022, https://arxiv.org/abs/2111.15430):
    'We train for 100 epochs with a learning rate of 0.1 for the first
    40 epochs, of 0.01 for the next 20 epochs and of 0.001 for the last
    40 epochs.'.
    """
    optimizer = optim.SGD(
        model.parameters(),
        lr=0.1,
        momentum=0.9,
        weight_decay=1e-4,
        nesterov=True,
    )
    scheduler = MultiStepLR(
        optimizer,
        milestones=[40, 60],
        gamma=0.1,
    )
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def optim_tinyimagenet_resnet50(
    model: nn.Module,
) -> dict[str, Optimizer | LRScheduler]:
    """Optimization recipe from 'The Devil is in the Margin: Margin-based
    Label Smoothing for Network Calibration',
    (CVPR 2022, https://arxiv.org/abs/2111.15430):
    'We train for 100 epochs with a learning rate of 0.1 for the first
    40 epochs, of 0.01 for the next 20 epochs and of 0.001 for the last
    40 epochs.'.
    """
    optimizer = optim.SGD(
        model.parameters(),
        lr=0.1,
        momentum=0.9,
        weight_decay=1e-4,
        nesterov=True,
    )
    scheduler = MultiStepLR(
        optimizer,
        milestones=[40, 60],
        gamma=0.1,
    )
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def batch_ensemble_wrapper(model: nn.Module, optim_recipe: Callable) -> dict:
    procedure = optim_recipe(model)
    param_optimizer = procedure["optimizer"]
    scheduler = procedure["lr_scheduler"]

    weight_decay = param_optimizer.defaults["weight_decay"]
    lr = param_optimizer.defaults["lr"]
    momentum = param_optimizer.defaults["momentum"]

    name_list = ["R", "S"]
    params_multi_tmp = list(
        filter(
            lambda kv: (name_list[0] in kv[0]) or (name_list[1] in kv[0]),
            model.named_parameters(),
        )
    )
    param_core_tmp = list(
        filter(
            lambda kv: (name_list[0] not in kv[0]) and (name_list[1] not in kv[0]),
            model.named_parameters(),
        )
    )

    params_multi = [param for _, param in params_multi_tmp]
    param_core = [param for _, param in param_core_tmp]
    optimizer = optim.SGD(
        [
            {"params": param_core, "weight_decay": weight_decay},
            {"params": params_multi, "weight_decay": 0.0},
        ],
        lr=lr,
        momentum=momentum,
    )

    scheduler.optimizer = optimizer
    return {"optimizer": optimizer, "lr_scheduler": scheduler}


def get_procedure(
    arch_name: str,
    ds_name: str,
    method: str = "",
    imagenet_recipe: str | None = None,
) -> Callable:
    """Get the optimization recipe for a given architecture and dataset.

    Args:
        arch_name (str): The name of the architecture.
        ds_name (str): The name of the dataset.
        method (str, optional): The name of the method. Defaults to "".
        imagenet_recipe (str, optional): The recipe to use for
            ImageNet. Defaults to None.

    Returns:
        callable: The optimization recipe.
    """
    if arch_name in ["resnet18", "resnet20"]:
        if ds_name == "cifar10":
            procedure = optim_cifar10_resnet18
        elif ds_name == "cifar100":
            procedure = optim_cifar100_resnet18
        else:
            raise NotImplementedError(f"Dataset {ds_name} not implemented.")
    elif arch_name == "resnet34":
        if ds_name == "cifar10":
            procedure = optim_cifar10_resnet34
        elif ds_name == "cifar100":
            procedure = optim_cifar100_resnet34
        elif ds_name == "tiny-imagenet":
            procedure = optim_tinyimagenet_resnet34
        else:
            raise NotImplementedError(f"No recipe for dataset: {ds_name}.")
    elif arch_name == "resnet50":
        if ds_name == "cifar10":
            procedure = optim_cifar10_resnet50
        elif ds_name == "cifar100":
            procedure = optim_cifar100_resnet50
        elif ds_name == "tiny-imagenet":
            procedure = optim_tinyimagenet_resnet50
        elif ds_name == "imagenet":
            if imagenet_recipe is not None and imagenet_recipe == "A3":
                procedure = optim_imagenet_resnet50_a3
            else:
                procedure = optim_imagenet_resnet50
        else:
            raise NotImplementedError(f"No recipe for dataset: {ds_name}.")
    elif arch_name == "wideresnet28x10":
        if ds_name in ("cifar10", "cifar100"):
            procedure = optim_cifar10_wideresnet
        else:
            raise NotImplementedError(f"No recipe for dataset: {ds_name}.")
    elif "vgg" in arch_name:
        if ds_name == "cifar10":
            procedure = optim_cifar10_vgg16
        elif ds_name == "cifar100":
            procedure = optim_cifar100_vgg16
        else:
            raise NotImplementedError(f"No recipe for dataset: {ds_name}.")
    else:
        raise NotImplementedError(f"No recipe for architecture: {arch_name}.")

    if method == "batched":
        procedure = partial(batch_ensemble_wrapper, optim_recipe=procedure)

    return procedure


class WarmupScheduler(SequentialLR):
    def __init__(
        self,
        optimizer: Optimizer,
        base_scheduler: type[LRScheduler],
        warmup_start_factor: float,
        warmup_epochs: int,
        scheduler_args: dict[str, float],
    ) -> None:
        """Scheduler with linear warmup.

        Args:
            optimizer (Optimizer): The optimizer to be used.*
            base_scheduler (type[LRScheduler]): The base scheduler class to use after
                the warmup.
            warmup_start_factor (float): The multiplicative factor to apply to
                the learning rate at the start of the warmup.
            warmup_epochs (int): The number of epochs to warmup the learning
                rate.
            scheduler_args (dict[str, float]): The arguments to pass to the base
                scheduler.
        """
        warmup_scheduler = LinearLR(
            optimizer,
            start_factor=warmup_start_factor,
            end_factor=1,
            total_iters=warmup_epochs,
        )
        base_scheduler = base_scheduler(optimizer, **scheduler_args)
        super().__init__(
            optimizer=optimizer,
            schedulers=[warmup_scheduler, base_scheduler],
            milestones=[warmup_epochs],
        )


class CosineAnnealingWarmup(WarmupScheduler):
    def __init__(
        self,
        optimizer: Optimizer,
        warmup_start_factor: float,
        warmup_epochs: int,
        max_epochs: int,
        eta_min: float = 0,
    ) -> None:
        """Cosine annealing scheduler with linear warmup.

        Args:
            optimizer (Optimizer): The optimizer to be used.
            warmup_start_factor (float): The multiplicative factor to apply to
                the learning rate at the start of the warmup.
            warmup_epochs (int): The number of epochs to warmup the learning
                rate.
            max_epochs (int): The total number of epochs including warmup.
            eta_min (float): The minimum learning rate.
        """
        super().__init__(
            optimizer=optimizer,
            base_scheduler=CosineAnnealingLR,
            warmup_start_factor=warmup_start_factor,
            warmup_epochs=warmup_epochs,
            scheduler_args={
                "T_max": max_epochs - warmup_epochs,
                "eta_min": eta_min,
            },
        )


class CosineSWALR(SequentialLR):
    def __init__(
        self,
        optimizer: Optimizer,
        milestone: int,
        swa_lr: float,
        anneal_epochs: int,
        optim_eta_min: float = 0,
        anneal_strategy: Literal["cos", "linear"] = "cos",
    ) -> None:
        """Chains a Cosine scheduler and a SWA scheduler.

        This class is an example of a wrapper to enable training SWA and SWAG
        models using the CLI. You may create your own class following this
        example.

        Args:
            optimizer (Optimizer): The optimizer to be used.
            milestone (int): The epoch to start the SWA.
            swa_lr (float): The learning rate to use for the SWA model.
            anneal_epochs (int): The number of epochs to anneal the learning rate.
            optim_eta_min (float): The minimum learning rate for the first optimizer.
            anneal_strategy (Literal["cos", "linear"]): The strategy to anneal the learning rate.
        """
        optim_scheduler = CosineAnnealingLR(
            optimizer=optimizer, T_max=milestone, eta_min=optim_eta_min
        )
        swa_scheduler = torch.optim.swa_utils.SWALR(
            optimizer,
            swa_lr=swa_lr,
            anneal_epochs=anneal_epochs,
            anneal_strategy=anneal_strategy,
        )
        super().__init__(
            optimizer=optimizer,
            schedulers=[optim_scheduler, swa_scheduler],
            milestones=[milestone],
        )



# ruff: noqa: F401
from .inverse import MeanAbsoluteErrorInverse, MeanSquaredErrorInverse
from .log10 import Log10
from .mse_log import MeanSquaredLogError
from .nll import DistributionNLL
from .relative_error import (
    MeanGTRelativeAbsoluteError,
    MeanGTRelativeSquaredError,
)
from .silog import SILog
from .threshold_accuracy import ThresholdAccuracy



import pytest
import torch

from torch_uncertainty.transforms import Mixup, MixupIO, RegMixup, WarpingMixup
from torch_uncertainty.transforms.mixup import AbstractMixup


@pytest.fixture()
def batch_input() -> tuple[torch.Tensor, torch.Tensor]:
    imgs = torch.rand(2, 3, 28, 28)
    return imgs, torch.tensor([0, 1])


class TestAbstractMixup:
    """Testing AbstractMixup augmentation."""

    def test_abstract_mixup(self, batch_input):
        with pytest.raises(NotImplementedError):
            AbstractMixup()(*batch_input)


class TestMixup:
    """Testing Mixup augmentation."""

    def test_batch_mixup(self, batch_input):
        mixup = Mixup(alpha=1.0, mode="batch", num_classes=2)
        _ = mixup(*batch_input)

    def test_elem_mixup(self, batch_input):
        mixup = Mixup(alpha=1.0, mode="elem", num_classes=2)
        _ = mixup(*batch_input)


class TestMixupIO:
    """Testing MixupIO augmentation."""

    def test_batch_mixupio(self, batch_input):
        mixup = MixupIO(alpha=1.0, mode="batch", num_classes=2)
        _ = mixup(*batch_input)

    def test_elem_mixupio(self, batch_input):
        mixup = MixupIO(alpha=1.0, mode="elem", num_classes=2)
        _ = mixup(*batch_input)


class TestRegMixup:
    """Testing RegMixup augmentation."""

    def test_batch_regmixup(self, batch_input):
        mixup = RegMixup(alpha=1.0, mode="batch", num_classes=2)
        _ = mixup(*batch_input)

    def test_elem_regmixup(self, batch_input):
        mixup = RegMixup(alpha=1.0, mode="elem", num_classes=2)
        _ = mixup(*batch_input)


class TestWarpingMixup:
    """Testing WarpingMixup augmentation."""

    def test_batch_kernel_warpingmixup(self, batch_input):
        mixup = WarpingMixup(alpha=1.0, mode="batch", num_classes=2, apply_kernel=True)
        _ = mixup(*batch_input, batch_input[0])

    def test_elem_kernel_warpingmixup(self, batch_input):
        mixup = WarpingMixup(alpha=1.0, mode="elem", num_classes=2, apply_kernel=True)
        _ = mixup(*batch_input, batch_input[0])

    def test_elem_warpingmixup(self, batch_input):
        mixup = WarpingMixup(alpha=1.0, mode="elem", num_classes=2, apply_kernel=False)
        _ = mixup(*batch_input, batch_input[0])



import logging
import math
from functools import partial

import torch
import torch.nn.functional as F
from timm.layers import DropPath, to_2tuple, trunc_normal_
from torch import Tensor, nn


class DWConv(nn.Module):
    def __init__(self, dim: int = 768) -> None:
        super().__init__()
        self.dwconv = nn.Conv2d(dim, dim, 3, 1, 1, bias=True, groups=dim)

    def forward(self, inputs: Tensor, h: int, w: int) -> Tensor:
        b, _, c = inputs.shape
        inputs = self.dwconv(inputs.transpose(1, 2).view(b, c, h, w))
        return inputs.flatten(2).transpose(1, 2)


class MLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: int | None = None,
        out_features: int | None = None,
        act_layer: type[nn.Module] = nn.GELU,
        dropout_rate: float = 0.0,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.dwconv = DWConv(hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.dropout = nn.Dropout(dropout_rate)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            nn.init.constant_(m.bias, 0)

    def forward(self, x, h, w):
        x = self.fc1(x)
        x = self.dwconv(x, h, w)
        x = self.act(x)
        x = self.dropout(x)
        x = self.fc2(x)
        return self.dropout(x)


class Attention(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int = 8,
        qkv_bias: bool = False,
        qk_scale: float | None = None,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
        sr_ratio: int = 1,
    ):
        super().__init__()
        assert dim % num_heads == 0, f"dim {dim} should be divided by num_heads {num_heads}."

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim**-0.5

        self.q = nn.Linear(dim, dim, bias=qkv_bias)
        self.kv = nn.Linear(dim, dim * 2, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.sr_ratio = sr_ratio
        if sr_ratio > 1:
            self.sr = nn.Conv2d(dim, dim, kernel_size=sr_ratio, stride=sr_ratio)
            self.norm = nn.LayerNorm(dim)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            nn.init.constant_(m.bias, 0)

    def forward(self, x: Tensor, h: int, w: int):
        b, n, c = x.shape
        q = self.q(x).reshape(b, n, self.num_heads, c // self.num_heads).permute(0, 2, 1, 3)

        if self.sr_ratio > 1:
            x_ = x.permute(0, 2, 1).reshape(b, c, h, w)
            x_ = self.sr(x_).reshape(b, c, -1).permute(0, 2, 1)
            x_ = self.norm(x_)
            kv = (
                self.kv(x_)
                .reshape(b, -1, 2, self.num_heads, c // self.num_heads)
                .permute(2, 0, 3, 1, 4)
            )
        else:
            kv = (
                self.kv(x)
                .reshape(b, -1, 2, self.num_heads, c // self.num_heads)
                .permute(2, 0, 3, 1, 4)
            )
        k, v = kv[0], kv[1]

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(b, n, c)
        x = self.proj(x)
        return self.proj_drop(x)


class Block(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = False,
        qk_scale: None | float = None,
        dropout: float = 0.0,
        attn_drop: float = 0.0,
        drop_path: float = 0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
        sr_ratio: int = 1,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            attn_drop=attn_drop,
            proj_drop=dropout,
            sr_ratio=sr_ratio,
        )
        # NOTE: drop path for stochastic depth, we shall see if this is better
        #   than dropout here
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = MLP(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            dropout_rate=dropout,
        )

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            nn.init.constant_(m.bias, 0)

    def forward(self, x, h, w):
        x = x + self.drop_path(self.attn(self.norm1(x), h, w))
        return x + self.drop_path(self.mlp(self.norm2(x), h, w))


class OverlapPatchEmbed(nn.Module):
    """Image to Patch Embedding."""

    def __init__(self, img_size=224, patch_size=7, stride=4, in_chans=3, embed_dim=768):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)

        self.img_size = img_size
        self.patch_size = patch_size
        self.h, self.w = (
            img_size[0] // patch_size[0],
            img_size[1] // patch_size[1],
        )
        self.num_patches = self.h * self.w
        self.proj = nn.Conv2d(
            in_chans,
            embed_dim,
            kernel_size=patch_size,
            stride=stride,
            padding=(patch_size[0] // 2, patch_size[1] // 2),
        )
        self.norm = nn.LayerNorm(embed_dim)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            nn.init.constant_(m.bias, 0)

    def forward(self, x):
        x = self.proj(x)
        _, _, h, w = x.shape
        x = x.flatten(2).transpose(1, 2)
        x = self.norm(x)

        return x, h, w


class MixVisionTransformer(nn.Module):
    def __init__(
        self,
        img_size: int,
        in_channels: int,
        num_classes: int,
        embed_dims: list[int],
        num_heads: list[int],
        mlp_ratios: list[int],
        qkv_bias: bool,
        qk_scale: float | None,
        drop_rate: float,
        attn_drop_rate: float,
        drop_path_rate: float,
        norm_layer,
        depths,
        sr_ratios: list[int],
    ):
        super().__init__()
        self.num_classes = num_classes
        self.depths = depths

        # patch_embed
        self.patch_embed1 = OverlapPatchEmbed(
            img_size=img_size,
            patch_size=7,
            stride=4,
            in_chans=in_channels,
            embed_dim=embed_dims[0],
        )
        self.patch_embed2 = OverlapPatchEmbed(
            img_size=img_size // 4,
            patch_size=3,
            stride=2,
            in_chans=embed_dims[0],
            embed_dim=embed_dims[1],
        )
        self.patch_embed3 = OverlapPatchEmbed(
            img_size=img_size // 8,
            patch_size=3,
            stride=2,
            in_chans=embed_dims[1],
            embed_dim=embed_dims[2],
        )
        self.patch_embed4 = OverlapPatchEmbed(
            img_size=img_size // 16,
            patch_size=3,
            stride=2,
            in_chans=embed_dims[2],
            embed_dim=embed_dims[3],
        )

        # transformer encoder
        dpr = [
            x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))
        ]  # stochastic depth decay rule
        cur = 0
        self.block1 = nn.ModuleList(
            [
                Block(
                    dim=embed_dims[0],
                    num_heads=num_heads[0],
                    mlp_ratio=mlp_ratios[0],
                    qkv_bias=qkv_bias,
                    qk_scale=qk_scale,
                    dropout=drop_rate,
                    attn_drop=attn_drop_rate,
                    drop_path=dpr[cur + i],
                    norm_layer=norm_layer,
                    sr_ratio=sr_ratios[0],
                )
                for i in range(depths[0])
            ]
        )
        self.norm1 = norm_layer(embed_dims[0])

        cur += depths[0]
        self.block2 = nn.ModuleList(
            [
                Block(
                    dim=embed_dims[1],
                    num_heads=num_heads[1],
                    mlp_ratio=mlp_ratios[1],
                    qkv_bias=qkv_bias,
                    qk_scale=qk_scale,
                    dropout=drop_rate,
                    attn_drop=attn_drop_rate,
                    drop_path=dpr[cur + i],
                    norm_layer=norm_layer,
                    sr_ratio=sr_ratios[1],
                )
                for i in range(depths[1])
            ]
        )
        self.norm2 = norm_layer(embed_dims[1])

        cur += depths[1]
        self.block3 = nn.ModuleList(
            [
                Block(
                    dim=embed_dims[2],
                    num_heads=num_heads[2],
                    mlp_ratio=mlp_ratios[2],
                    qkv_bias=qkv_bias,
                    qk_scale=qk_scale,
                    dropout=drop_rate,
                    attn_drop=attn_drop_rate,
                    drop_path=dpr[cur + i],
                    norm_layer=norm_layer,
                    sr_ratio=sr_ratios[2],
                )
                for i in range(depths[2])
            ]
        )
        self.norm3 = norm_layer(embed_dims[2])

        cur += depths[2]
        self.block4 = nn.ModuleList(
            [
                Block(
                    dim=embed_dims[3],
                    num_heads=num_heads[3],
                    mlp_ratio=mlp_ratios[3],
                    qkv_bias=qkv_bias,
                    qk_scale=qk_scale,
                    dropout=drop_rate,
                    attn_drop=attn_drop_rate,
                    drop_path=dpr[cur + i],
                    norm_layer=norm_layer,
                    sr_ratio=sr_ratios[3],
                )
                for i in range(depths[3])
            ]
        )
        self.norm4 = norm_layer(embed_dims[3])

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            nn.init.constant_(m.bias, 0)

    def forward_features(self, x):
        b = x.shape[0]
        outs = []

        # stage 1
        x, h, w = self.patch_embed1(x)
        for blk in self.block1:
            x = blk(x, h, w)
        x = self.norm1(x)
        x = x.reshape(b, h, w, -1).permute(0, 3, 1, 2).contiguous()
        outs.append(x)

        # stage 2
        x, h, w = self.patch_embed2(x)
        for blk in self.block2:
            x = blk(x, h, w)
        x = self.norm2(x)
        x = x.reshape(b, h, w, -1).permute(0, 3, 1, 2).contiguous()
        outs.append(x)

        # stage 3
        x, h, w = self.patch_embed3(x)
        for blk in self.block3:
            x = blk(x, h, w)
        x = self.norm3(x)
        x = x.reshape(b, h, w, -1).permute(0, 3, 1, 2).contiguous()
        outs.append(x)

        # stage 4
        x, h, w = self.patch_embed4(x)
        for blk in self.block4:
            x = blk(x, h, w)
        x = self.norm4(x)
        x = x.reshape(b, h, w, -1).permute(0, 3, 1, 2).contiguous()
        outs.append(x)

        return outs

    def forward(self, x):
        return self.forward_features(x)


def _get_embed_dims(arch: int) -> list[int]:
    if arch == 0:
        return [32, 64, 160, 256]
    return [64, 128, 320, 512]


def _get_depths(arch: int) -> list[int]:  # coverage: ignore
    if arch == 0 or arch == 1:
        return [2, 2, 2, 2]
    if arch == 2:
        return [3, 4, 6, 3]
    if arch == 3:
        return [3, 4, 18, 3]
    if arch == 4:
        return [3, 8, 27, 3]
    # arch == 5:
    return [3, 6, 40, 3]


class Mit(MixVisionTransformer):
    def __init__(self, arch: int):
        embed_dims = _get_embed_dims(arch)
        depths = _get_depths(arch)
        super().__init__(
            img_size=224,
            in_channels=3,
            num_classes=1000,
            qk_scale=None,
            embed_dims=embed_dims,
            num_heads=[1, 2, 5, 8],
            mlp_ratios=[4, 4, 4, 4],
            qkv_bias=True,
            norm_layer=partial(nn.LayerNorm, eps=1e-6),
            depths=depths,
            sr_ratios=[8, 4, 2, 1],
            drop_rate=0.0,
            drop_path_rate=0.1,
            attn_drop_rate=0.0,
        )


class MLPHead(nn.Module):
    """Linear Embedding with transposition."""

    def __init__(self, input_dim: int = 2048, embed_dim: int = 768) -> None:
        super().__init__()
        self.proj = nn.Linear(input_dim, embed_dim)

    def forward(self, inputs: Tensor) -> Tensor:
        return self.proj(inputs.flatten(2).transpose(1, 2))


def resize(
    inputs: Tensor,
    size: torch.Size | None = None,
    scale_factor=None,
    mode: str = "nearest",
    align_corners: bool | None = None,
    warning: bool = True,
) -> Tensor:
    if warning and size is not None and align_corners:  # coverage: ignore
        input_h, input_w = tuple(int(x) for x in inputs.shape[2:])
        output_h, output_w = tuple(int(x) for x in size)
        if (output_h > input_h or output_w > output_h) and (
            (output_h > 1 and output_w > 1 and input_h > 1 and input_w > 1)
            and (output_h - 1) % (input_h - 1)
            and (output_w - 1) % (input_w - 1)
        ):
            logging.info(
                "When align_corners=%s, "
                "the output would more aligned if "
                "input size %s is `x+1` and "
                "out size %s is `nx+1`",
                align_corners,
                (input_h, input_w),
                (output_h, output_w),
            )
    return F.interpolate(inputs, size, scale_factor, mode, align_corners)


class SegFormerHead(nn.Module):
    """Head for SegFormer.

    Reference:
        SegFormer: Simple and Efficient Design for Semantic Segmentation with
        Transformers.
    """

    def __init__(
        self,
        in_channels: list[int],
        feature_strides: list[int],
        embed_dim: int,
        num_classes: int,
        dropout_ratio: float = 0.1,
    ):
        super().__init__()
        self.in_channels = in_channels
        assert len(feature_strides) == len(in_channels)
        assert min(feature_strides) == feature_strides[0]

        self.linear_c4 = MLPHead(input_dim=in_channels[3], embed_dim=embed_dim)
        self.linear_c3 = MLPHead(input_dim=in_channels[2], embed_dim=embed_dim)
        self.linear_c2 = MLPHead(input_dim=in_channels[1], embed_dim=embed_dim)
        self.linear_c1 = MLPHead(input_dim=in_channels[0], embed_dim=embed_dim)

        self.fuse = nn.Sequential(
            nn.Conv2d(embed_dim * 4, embed_dim, kernel_size=1, bias=False),
            nn.ReLU(),
            nn.BatchNorm2d(embed_dim),
        )
        self.classifier = nn.Conv2d(embed_dim, num_classes, kernel_size=1)
        self.dropout = nn.Dropout2d(dropout_ratio)

    def forward(self, inputs: Tensor) -> Tensor:
        # x [inputs[i] for i in self.in_index] # len=4, 1/4,1/8,1/16,1/32
        c1, c2, c3, c4 = inputs[0], inputs[1], inputs[2], inputs[3]

        n, _, _, _ = c4.shape

        _c4 = self.linear_c4(c4).permute(0, 2, 1).reshape(n, -1, c4.shape[2], c4.shape[3])
        _c4 = resize(_c4, size=c1.size()[2:], mode="bilinear", align_corners=False)

        _c3 = self.linear_c3(c3).permute(0, 2, 1).reshape(n, -1, c3.shape[2], c3.shape[3])
        _c3 = resize(_c3, size=c1.size()[2:], mode="bilinear", align_corners=False)

        _c2 = self.linear_c2(c2).permute(0, 2, 1).reshape(n, -1, c2.shape[2], c2.shape[3])
        _c2 = resize(_c2, size=c1.size()[2:], mode="bilinear", align_corners=False)

        _c1 = self.linear_c1(c1).permute(0, 2, 1).reshape(n, -1, c1.shape[2], c1.shape[3])

        _c = self.fuse(torch.cat([_c4, _c3, _c2, _c1], dim=1))

        x = self.dropout(_c)
        return self.classifier(x)


class _SegFormer(nn.Module):
    def __init__(
        self,
        in_channels: list[int],
        feature_strides: list[int],
        embed_dim: int,
        num_classes: int,
        dropout_ratio: float,
        mit: nn.Module,
    ):
        super().__init__()

        self.encoder = mit
        self.head = SegFormerHead(
            in_channels,
            feature_strides,
            embed_dim,
            num_classes,
            dropout_ratio,
        )

    def forward(self, inputs: Tensor) -> Tensor:
        features = self.encoder(inputs)
        return self.head(features)


def seg_former(num_classes: int, arch: int) -> _SegFormer:
    in_channels = _get_embed_dims(arch)
    return _SegFormer(
        in_channels=in_channels,
        feature_strides=[4, 8, 16, 32],
        embed_dim=256 if arch == 0 else 512,
        num_classes=num_classes,
        dropout_ratio=0.1,
        mit=Mit(arch),
    )



# ruff: noqa: F401
from .classification import (
    AUGRC,
    AURC,
    FPR95,
    AdaptiveCalibrationError,
    BrierScore,
    CalibrationError,
    CategoricalNLL,
    CovAt5Risk,
    CovAtxRisk,
    Disagreement,
    Entropy,
    GroupingLoss,
    MeanIntersectionOverUnion,
    MutualInformation,
    RiskAt80Cov,
    VariationRatio,
)
from .regression import (
    DistributionNLL,
    Log10,
    MeanAbsoluteErrorInverse,
    MeanGTRelativeAbsoluteError,
    MeanGTRelativeSquaredError,
    MeanSquaredErrorInverse,
    MeanSquaredLogError,
    SILog,
    ThresholdAccuracy,
)
from .sparsification import AUSE



# ruff: noqa: F401, F403
from .batched import batched_resnet
from .lpbnn import lpbnn_resnet
from .masked import masked_resnet
from .mimo import mimo_resnet
from .packed import packed_resnet
from .std import resnet



import logging
from collections.abc import Callable
from pathlib import Path
from typing import Any

from PIL import Image
from torchvision.datasets import VisionDataset
from torchvision.datasets.utils import (
    check_integrity,
    download_and_extract_archive,
)


def pil_loader(path: Path) -> Image.Image:
    # open path as file to avoid ResourceWarning (https://github.com/python-pillow/Pillow/issues/835)
    with path.open("rb") as f:
        img = Image.open(f)
        return img.convert("RGB")


class FrostImages(VisionDataset):  # TODO: Use ImageFolder
    url = "https://zenodo.org/records/10438904/files/frost.zip"
    zip_md5 = "d82f29f620d43a68e71e34b28f7c35cb"
    filename = "frost.zip"
    samples = [
        "frost1.png",
        "frost2.png",
        "frost3.jpg",
        "frost4.jpg",
        "frost5.jpg",
    ]

    def __init__(
        self,
        root: str | Path,
        transform: Callable[..., Any] | None,
        target_transform: Callable[..., Any] | None = None,
        download: bool = False,
    ) -> None:
        self.root = Path(root)

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted. You can use download=True to download it."
            )

        super().__init__(
            self.root / "frost",
            transform=transform,
            target_transform=target_transform,
        )
        self.loader = pil_loader

    def _check_integrity(self) -> bool:
        fpath = self.root / self.filename
        return check_integrity(
            fpath,
            self.zip_md5,
        )

    def download(self) -> None:
        if self._check_integrity():
            logging.info("Files already downloaded and verified")
            return

        download_and_extract_archive(
            self.url,
            download_root=self.root,
            filename=self.filename,
            md5=self.zip_md5,
        )
        logging.info("Downloaded %s to %s.", self.filename, self.root)

    def __getitem__(self, index: int) -> Any:
        """Get the samples of the dataset.

        Args:
            index (int): Index

        Returns:
            tuple: (sample, target) where target is class_index of the target class.
        """
        path = self.root / self.samples[index]
        sample = self.loader(path)
        if self.transform is not None:
            sample = self.transform(sample)
        return sample

    def __len__(self) -> int:
        """Get the length of the dataset."""
        return len(self.samples)



from typing import Literal

import torch
import torchvision.models as tv_models
from torch import Tensor, nn
from torch.nn import functional as F
from torch.nn.common_types import _size_2_t
from torchvision.models.resnet import ResNet50_Weights, ResNet101_Weights

from torch_uncertainty.models.utils import Backbone, set_bn_momentum


class SeparableConv2d(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_2_t,
        stride: _size_2_t = 1,
        padding: _size_2_t = 0,
        dilation: _size_2_t = 1,
        bias=True,
    ) -> None:
        """Separable Convolution with dilation.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            kernel_size (_size_2_t): Kernel size.
            stride (_size_2_t, optional): Stride. Defaults to 1.
            padding (_size_2_t, optional): Padding. Defaults to 0.
            dilation (_size_2_t, optional): Dilation. Defaults to 1.
            bias (bool, optional): Use biases. Defaults to True.
        """
        super().__init__()
        self.separable = nn.Conv2d(
            in_channels,
            in_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            bias=bias,
            groups=in_channels,
        )
        self.pointwise = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=bias,
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.pointwise(self.separable(x))


class InnerConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        dilation: _size_2_t,
        separable: bool,
    ) -> None:
        """Inner convolution block.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            dilation (_size_2_t): Dilation.
            separable (bool): Use separable convolutions to reduce the number
                of parameters.
        """
        super().__init__()
        if not separable:
            self.conv = nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=3,
                padding=dilation,
                dilation=dilation,
                bias=False,
            )
        else:
            self.conv = SeparableConv2d(
                in_channels,
                out_channels,
                kernel_size=3,
                padding=dilation,
                dilation=dilation,
                bias=False,
            )
        self.bn = nn.BatchNorm2d(out_channels)

    def forward(self, x: Tensor) -> Tensor:
        return F.relu(self.bn(self.conv(x)))


class InnerPooling(nn.Module):
    def __init__(self, in_channels: int, out_channels: int) -> None:
        """Inner pooling block.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
        """
        super().__init__()
        self.pool = nn.AdaptiveAvgPool2d(1)
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)
        self.bn = nn.BatchNorm2d(out_channels)

    def forward(self, x: Tensor) -> Tensor:
        size = x.shape[-2:]
        x = F.relu(self.bn(self.conv(self.pool(x))))
        return F.interpolate(x, size=size, mode="bilinear", align_corners=False)


class ASPP(nn.Module):
    def __init__(
        self,
        in_channels: int,
        atrous_rates: list[int],
        separable: bool,
        dropout_rate: float,
    ) -> None:
        """Atrous Spatial Pyramid Pooling.

        Args:
            in_channels (int): Number of input channels.
            atrous_rates (list[int]): Atrous rates for the ASPP module.
            separable (bool): Use separable convolutions to reduce the number
                of parameters.
            dropout_rate (float): Dropout rate of the ASPP.
        """
        super().__init__()
        out_channels = 256
        modules = []
        modules.append(
            nn.Sequential(
                nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False),
                nn.BatchNorm2d(out_channels),
                nn.ReLU(inplace=True),
            )
        )
        modules += [
            InnerConv(in_channels, out_channels, dilation, separable) for dilation in atrous_rates
        ]
        modules.append(InnerPooling(in_channels, out_channels))
        self.convs = nn.ModuleList(modules)

        self.projection = nn.Sequential(
            nn.Conv2d(5 * out_channels, out_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Dropout(dropout_rate),
        )

    def forward(self, x: Tensor) -> Tensor:
        res = torch.cat([conv(x) for conv in self.convs], dim=1)
        return self.projection(res)


class DeepLabV3Backbone(Backbone):
    def __init__(
        self,
        backbone_name: Literal["resnet50", "resnet101"],
        style: str,
        pretrained: bool,
        norm_momentum: float,
    ) -> None:
        """DeepLab V3(+) backbone.

        Args:
            backbone_name (str): Backbone name.
            style (str): Whether to use a DeepLab V3 or V3+ model.
            pretrained (bool): Use pretrained backbone.
            norm_momentum (float): BatchNorm momentum.
        """
        # TODO: handle dilations
        if backbone_name == "resnet50":
            base_model = tv_models.resnet50(
                weights=ResNet50_Weights.DEFAULT if pretrained else None
            )
        elif backbone_name == "resnet101":
            base_model = tv_models.resnet101(
                weights=ResNet101_Weights.DEFAULT if pretrained else None
            )
        else:
            raise ValueError(f"Unknown backbone: {backbone_name}.")
        base_model.avgpool = nn.Identity()
        base_model.fc = nn.Identity()
        set_bn_momentum(base_model, norm_momentum)

        feat_names = ["layer1", "layer4"] if style == "v3+" else ["layer4"]
        super().__init__(base_model, feat_names)


class DeepLabV3Decoder(nn.Module):
    """Decoder for the DeepLabV3 model.

    Args:
        in_channels (int): Number of channels of the input latent space.
        num_classes (int): Number of classes.
        aspp_dilate (list[int], optional): Atrous rates for the ASPP module.
        separable (bool, optional): Use separable convolutions to reduce the number
            of parameters. Defaults to False.
        dropout_rate (float, optional): Dropout rate of the ASPP. Defaults to 0.1.
    """

    def __init__(
        self,
        in_channels: int,
        num_classes: int,
        aspp_dilate: list[int],
        separable: bool = False,
        dropout_rate: float = 0.1,
    ) -> None:
        super().__init__()
        self.aspp = ASPP(in_channels, aspp_dilate, separable, dropout_rate)
        if not separable:
            self.conv = nn.Conv2d(256, 256, 3, padding=1, bias=False)
        else:
            self.conv = SeparableConv2d(256, 256, 3, padding=1, bias=False)
        self.bn = nn.BatchNorm2d(256)
        self.classifier = nn.Conv2d(256, num_classes, kernel_size=1)

    def forward(self, features: list[Tensor]) -> Tensor:
        out = F.relu(self.bn(self.conv(self.aspp(features[0]))))
        return self.classifier(out)


class DeepLabV3PlusDecoder(nn.Module):
    def __init__(
        self,
        in_channels: int,
        low_level_channels: int,
        num_classes: int,
        aspp_dilate: list[int],
        separable: bool,
        dropout_rate: float = 0.1,
    ) -> None:
        """Decoder for the DeepLabV3+ model.

        Args:
            in_channels (int): Number of channels of the input latent space.
            low_level_channels (int): Number of low-level features channels.
            num_classes (int): Number of classes.
            aspp_dilate (list[int]): Atrous rates for the ASPP module.
            separable (bool): Use separable convolutions to reduce the number
                of parameters.
            dropout_rate (float, optional): Dropout rate of the ASPP. Defaults
                to 0.1.
        """
        super().__init__()
        self.project = nn.Sequential(
            nn.Conv2d(low_level_channels, 48, kernel_size=1, bias=False),
            nn.BatchNorm2d(48),
            nn.ReLU(inplace=True),
        )
        self.atrous_spatial_pyramid_pool = ASPP(in_channels, aspp_dilate, separable, dropout_rate)
        if separable:
            self.conv = SeparableConv2d(304, 256, 3, padding=1, bias=False)
        else:
            self.conv = nn.Conv2d(304, 256, 3, padding=1, bias=False)
        self.bn = nn.BatchNorm2d(256)
        self.classifier = nn.Conv2d(256, num_classes, kernel_size=1)

    def forward(self, features: list[Tensor]) -> Tensor:
        low_level_features = self.project(features[0])
        output_features = self.atrous_spatial_pyramid_pool(features[1])
        output_features = F.interpolate(
            output_features,
            size=low_level_features.shape[2:],
            mode="bilinear",
            align_corners=False,
        )
        output_features = torch.cat([low_level_features, output_features], dim=1)
        out = F.relu(self.bn(self.conv(output_features)))
        return self.classifier(out)


class _DeepLabV3(nn.Module):
    def __init__(
        self,
        num_classes: int,
        backbone_name: str,
        style: Literal["v3", "v3+"],
        output_stride: int = 16,
        separable: bool = False,
        pretrained_backbone: bool = True,
        norm_momentum: float = 0.01,
    ) -> None:
        """DeepLab V3(+) model.

        Args:
            num_classes (int): Number of classes.
            backbone_name (str): Backbone name.
            style (Literal["v3", "v3+"]):  Whether to use a DeepLab V3 or
                V3+ model.
            output_stride (int, optional): Output stride. Defaults to 16.
            separable (bool, optional): Use separable convolutions. Defaults
                to False.
            pretrained_backbone (bool, optional): Use pretrained backbone.
                Defaults to True.
            norm_momentum (float, optional): BatchNorm momentum. Defaults to
                0.01.

        References:
            - Rethinking atrous convolution for semantic image segmentation.
            Chen, L. C., Papandreou, G., Schroff, F., & Adam, H. (2018).
            - Encoder-decoder with atrous separable convolution for semantic image segmentation.
            Chen, L. C., Zhu, Y., Papandreou, G., Schroff, F., & Adam, H. In ECCV 2018.
        """
        super().__init__()
        if output_stride == 16:
            dilations = [6, 12, 18]
        elif output_stride == 8:
            dilations = [12, 24, 36]
        else:
            raise ValueError(f"output_stride: {output_stride} is not supported.")

        self.backbone = DeepLabV3Backbone(backbone_name, style, pretrained_backbone, norm_momentum)
        if style == "v3":
            self.decoder = DeepLabV3Decoder(
                in_channels=2048,
                num_classes=num_classes,
                aspp_dilate=dilations,
                separable=separable,
                dropout_rate=0.1,
            )
        elif style == "v3+":
            self.decoder = DeepLabV3PlusDecoder(
                in_channels=2048,
                low_level_channels=256,
                num_classes=num_classes,
                aspp_dilate=dilations,
                separable=separable,
                dropout_rate=0.1,
            )
        else:
            raise ValueError(f"Unknown style: {style}.")

    def forward(self, x: Tensor) -> Tensor:
        input_shape = x.shape[-2:]
        return F.interpolate(
            self.decoder(self.backbone(x)),
            size=input_shape,
            mode="bilinear",
            align_corners=False,
        )


def deep_lab_v3_resnet50(
    num_classes: int,
    style: Literal["v3", "v3+"],
    output_stride: int = 16,
    separable: bool = False,
    pretrained_backbone: bool = True,
) -> _DeepLabV3:
    """DeepLab V3(+) model with ResNet-50 backbone.

    Args:
        num_classes (int): Number of classes.
        style (Literal["v3", "v3+"]): Whether to use a DeepLab V3 or V3+ model.
        output_stride (int, optional): Output stride. Defaults to 16.
        separable (bool, optional): Use separable convolutions. Defaults to
            False.
        pretrained_backbone (bool, optional): Use pretrained backbone. Defaults
            to True.
    """
    return _DeepLabV3(
        num_classes,
        "resnet50",
        style,
        output_stride=output_stride,
        separable=separable,
        pretrained_backbone=pretrained_backbone,
    )


def deep_lab_v3_resnet101(
    num_classes: int,
    style: Literal["v3", "v3+"],
    output_stride: int = 16,
    separable: bool = False,
    pretrained_backbone: bool = True,
) -> _DeepLabV3:
    """DeepLab V3(+) model with ResNet-50 backbone.

    Args:
        num_classes (int): Number of classes.
        style (Literal["v3", "v3+"]): Whether to use a DeepLab V3 or V3+ model.
        output_stride (int, optional): Output stride. Defaults to 16.
        separable (bool, optional): Use separable convolutions. Defaults to False.
        pretrained_backbone (bool, optional): Use pretrained backbone. Defaults to True.
    """
    return _DeepLabV3(
        num_classes,
        "resnet101",
        style,
        output_stride=output_stride,
        separable=separable,
        pretrained_backbone=pretrained_backbone,
    )



import logging
from collections.abc import Callable
from importlib import util
from pathlib import Path

if util.find_spec("pandas"):
    import pandas as pd

    pandas_installed = True
else:  # coverage: ignore
    pandas_installed = False


import torch
import torch.nn.functional as F
from torch.utils.data import Dataset
from torchvision.datasets.utils import (
    check_integrity,
    download_and_extract_archive,
    download_url,
)

boston_column_names = [
    "CRIM",
    "ZN",
    "INDUS",
    "CHAS",
    "NOX",
    "RM",
    "AGE",
    "DIS",
    "RAD",
    "TAX",
    "PTRATIO",
    "B",
    "LSTAT",
    "MEDV",
]

energy_prediction_column_names = [
    "Appliances",
    "lights",
    "T1",
    "RH_1",
    "T2",
    "RH_2",
    "T3",
    "RH_3",
    "T4",
    "RH_4",
    "T5",
    "RH_5",
    "T6",
    "RH_6",
    "T7",
    "RH_7",
    "T8",
    "RH_8",
    "T9",
    "RH_9",
    "T_out",  # Dropped
]


class UCIRegression(Dataset):
    """The UCI regression datasets.

    Args:
        root (str): Root directory of the datasets.
        train (bool, optional): If True, creates dataset from training set,
            otherwise creates from test set.
        transform (callable, optional): A function/transform that takes in a
            numpy array and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes
            in the target and transforms it.
        dataset_name (str, optional): The name of the dataset. One of
            "boston-housing", "concrete", "energy", "kin8nm",
            "naval-propulsion-plant", "power-plant", "protein",
            "wine-quality-red", and "yacht".
        download (bool, optional): If true, downloads the dataset from the
            internet and puts it in root directory. If dataset is already
            downloaded, it is not downloaded again.

    Note - Ethics:
        You may want to avoid using the boston-housing dataset because of
        ethical concerns.

    Note - License:
        The licenses of the datasets may differ from TorchUncertainty's
        license. Check before use.
    """

    root_appendix = "uci_regression"
    uci_subsets = [
        "boston",
        "concrete",
        "energy-efficiency",
        "energy-prediction",
        "kin8nm",
        "naval-propulsion-plant",
        "power-plant",
        "protein",
        "wine-quality-red",
        "yacht",
    ]

    md5_tgz = [
        "d4accdce7a25600298819f8e28e8d593",
        "eba3e28907d4515244165b6b2c311b7b",
        "2018fb7b50778fdc1304d50a78874579",
        "d0f0f8ceaaf45df2233ce0600097bd84",
        "df08c665b7665809e74e32b107836a3a",
        "54f4febcf51bdba12e1ca63e28b3e973",
        "f5065a616eae05eb4ecae445ecf6e720",
        "37bcb77a8abad274a987439e6a3de632",
        "0ddfa7a9379510fe7ff88b9930e3c332",
        "4e6727f462779e2d396e8f7d2ddb79a3",
    ]
    urls = [
        "https://archive.ics.uci.edu/ml/machine-learning-databases/housing/housing.data",
        "https://archive.ics.uci.edu/static/public/165/concrete+compressive+strength.zip",
        "https://archive.ics.uci.edu/static/public/242/energy+efficiency.zip",
        "https://archive.ics.uci.edu/static/public/374/appliances+energy+prediction.zip",
        "https://zenodo.org/records/14645866/files/kin8nm.csv",
        "https://raw.githubusercontent.com/luishpinto/cm-naval-propulsion-plant/master/data.csv",
        "https://archive.ics.uci.edu/static/public/294/combined+cycle+power+plant.zip",
        "https://archive.ics.uci.edu/static/public/265/physicochemical+"
        "properties+of+protein+tertiary+structure.zip",
        "https://archive.ics.uci.edu/static/public/186/wine+quality.zip",
        "https://archive.ics.uci.edu/static/public/243/yacht+hydrodynamics.zip",
    ]

    def __init__(
        self,
        root: Path | str,
        transform: Callable | None = None,
        target_transform: Callable | None = None,
        dataset_name: str = "energy",
        download: bool = False,
        seed: int = 42,
        shuffle: bool = True,
    ) -> None:
        super().__init__()
        self.root = Path(root)
        self.transform = transform
        self.target_transform = target_transform
        self.seed = seed
        self.shuffle = shuffle

        if dataset_name not in self.uci_subsets:
            raise ValueError(
                f"The dataset {dataset_name} is not implemented. "
                "`dataset_name` should be one of {self.uci_subsets}."
            )
        self.dataset_name = dataset_name
        dataset_id = self.uci_subsets.index(dataset_name)
        self.url = self.urls[dataset_id]
        self.start_filename = self.url.split("/")[-1]
        self.md5 = self.md5_tgz[dataset_id]

        if download:
            self.download()

        self._make_dataset()

    def __len__(self) -> int:
        """Get the length of the dataset."""
        return self.data.shape[0]

    def _check_integrity(self) -> bool:
        """Check the integrity of the dataset(s)."""
        return check_integrity(
            self.root / self.root_appendix / Path(self.start_filename),
            self.md5,
        )

    def _standardize(self) -> None:
        self.data = (self.data - self.data_mean) / self.data_std
        self.targets = (self.targets - self.target_mean) / self.target_std

    def _compute_statistics(self) -> None:
        self.data_mean = self.data.mean(axis=0)
        self.data_std = self.data.std(axis=0)
        self.data_std[self.data_std == 0] = 1
        self.target_mean = self.targets.mean(axis=0)
        self.target_std = self.targets.std(axis=0)

    def download(self) -> None:
        """Download and extract dataset."""
        if self._check_integrity():
            logging.info("Files already downloaded and verified")
            return
        if self.url is None:
            raise ValueError(f"The dataset {self.dataset_name} is not available for download.")
        download_root = self.root / self.root_appendix / self.dataset_name
        if self.dataset_name == "boston":
            download_url(
                self.url,
                root=download_root,
                filename="housing.data",
            )
        elif self.dataset_name == "kin8nm":
            download_url(
                self.url,
                root=download_root,
                filename="kin8nm.csv",
            )
        elif self.dataset_name == "naval-propulsion-plant":
            download_url(
                self.url,
                root=download_root,
                filename="data.csv",
            )
        else:
            download_and_extract_archive(
                self.url,
                download_root=download_root,
                extract_root=download_root,
                filename=self.start_filename,
                md5=self.md5,
            )

    def _make_dataset(self) -> None:
        """Create dataset from extracted files."""
        if not pandas_installed:  # coverage: ignore
            raise ImportError(
                "Please install torch_uncertainty with the tabular option:"
                """pip install -U "torch_uncertainty[tabular]"."""
            )
        path = self.root / self.root_appendix / self.dataset_name
        if self.dataset_name == "boston":
            array = pd.read_table(
                path / "housing.data",
                names=boston_column_names,
                header=None,
                delim_whitespace=True,
            )
        elif self.dataset_name == "concrete":
            array = pd.read_excel(path / "Concrete_Data.xls").to_numpy()
        elif self.dataset_name == "energy-efficiency":
            array = pd.read_excel(path / "ENB2012_data.xlsx").to_numpy()
        elif self.dataset_name == "energy-prediction":
            array = pd.read_csv(path / "energydata_complete.csv")[
                energy_prediction_column_names
            ].to_numpy()
        elif self.dataset_name == "kin8nm":
            array = pd.read_csv(path / "kin8nm.csv").to_numpy()
        elif self.dataset_name == "naval-propulsion-plant":
            df = pd.read_csv(path / "data.csv", header=None, sep=";", decimal=",")
            # convert Ex to 10^x and remove second target
            array = df.apply(pd.to_numeric, errors="coerce").to_numpy()[:, :-1]
        elif self.dataset_name == "protein":
            array = pd.read_csv(
                path / "CASP.csv",
            ).to_numpy()
        elif self.dataset_name == "wine-quality-red":
            array = pd.read_csv(
                path / "winequality-red.csv",
                sep=";",
            ).to_numpy()
        elif self.dataset_name == "yacht":
            array = pd.read_csv(
                path / "yacht_hydrodynamics.data",
                delim_whitespace=True,
                header=None,
            ).to_numpy()
        else:
            raise ValueError("Dataset not implemented.")

        array = torch.as_tensor(array).float()

        if self.dataset_name == "energy-efficiency":
            self.data = array[:, 2:-3]
            self.targets = array[:, -2]
        else:
            self.data = array[:, :-1]
            self.targets = array[:, -1]

        self._compute_statistics()
        self._standardize()

        if self.dataset_name == "energy-prediction":
            self.data = F.pad(self.data, (0, 0, 13, 0), value=0)

        if self.shuffle:
            gen = torch.Generator()
            gen.manual_seed(self.seed)
            indexes = torch.randperm(array.shape[0], generator=gen)
            array = array[indexes]

    def __getitem__(self, index: int) -> tuple[torch.Tensor, torch.Tensor]:
        """Get sample and target for a given index."""
        if self.dataset_name == "energy-prediction":
            data = self.data[index : index + 13, :]
            target = self.data[index : index + 13, :]
            return data, target

        data = self.data[index]
        if self.transform is not None:
            data = self.transform(data)
        target = self.targets[index]
        if self.target_transform is not None:
            target = self.target_transform(target)
        return data, target
